/*
 * CACHEUTL.C - Module to implement cached file I/O for performance-enhancement
 *
 * This module implements file I/O using a homebrew cache system to
 * avoid lots of small reads & writes through the Windows / DOS
 * file system (seems to work pretty well on Unix, too)!
 *
 * Copyright (C) 1997 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#include "lcadrfi.h"

#if SENTINEL_SIZE
SPROTO(void, SentinelTest, (PCacheControl pCache));

STATIC_FN(void) SentinelTest(pCache)PCacheControl pCache; {
    if (pCache) {
        TUShort i;

        for ( i = 0; i < SENTINEL_SIZE; ++i ) {
            if (pCache->preSentinel[i] != (char)0xE5) {
                return;
            }

            if (pCache->postSentinel[i] != (char)0xE5) {
                return;
            }
        }
    }
} /* end of: SentinelTest */

#else

#define SentinelTest(arg)

#endif

#ifdef _DEBUG

SPROTO(PCacheControl,   CacheAlloc,     (TUShort cacheSize,
                         LPCSTR  pFileName));
#else

SPROTO(PCacheControl, CacheAlloc, (TUShort cacheSize));

#endif

SPROTO(void, CacheFree, (PCacheControl * ppCache));

SPROTO(void, CacheClean, (PCacheControl pCache));

SPROTO(void, CacheFill, (PCacheControl pCache));

PCacheControl CacheOpen(pFileName, mode, cacheSize, errSeek, errRead, errWrite, errOpen, callFnOnErr) LPCSTR pFileName;
    LPCSTR mode;
    TUShort cacheSize;
    TErrCode errSeek;
    TErrCode errRead;
    TErrCode errWrite;
    TErrCode errOpen;
    TBoolByte callFnOnErr; {
    /* Local variables */
    TBoolByte writeable = FALSE;
    PCacheControl pCache = NULL;
    const char* p = mode;

    if (*p++ == 'w') {
        writeable = TRUE;
    } else if (*p == '+') {
        writeable = TRUE;
    }

    if ((cacheSize == 0) || (cacheSize > MAX_CACHE_SIZE)) {
        cacheSize = MAX_CACHE_SIZE;
    }
#ifdef _DEBUG

    pCache = CacheAlloc(cacheSize, pFileName);

#else

    pCache = CacheAlloc(cacheSize);

#endif

    if (pCache) {
        if (!writeable) {
            pCache->pFile = fopen(pFileName, mode);
        } else {
            pCache->pFile = FSOPEN(pFileName, mode, _SH_DENYWR);
        }

        if (!(pCache->pFile)) {
            if (callFnOnErr) {
                CacheError(errOpen);
            }
            CacheFree(&pCache);
        } else if (fseek(pCache->pFile, 0, SEEK_END)) {
            if (callFnOnErr) {
                CacheError(errSeek);
            }
            fclose(pCache->pFile);
            CacheFree(&pCache);
        } else {
            pCache->fileSize = ftell(pCache->pFile);
            pCache->curFileAddrOffset = 0;
            pCache->nextRWAddr = 0;
            pCache->cacheStartAddr = CACHE_START_DEACTIVATE;
            pCache->cacheEndAddr = 0;
            pCache->cacheErrStatus = 0;
            pCache->cacheErrSeek = errSeek;
            pCache->cacheErrRead = errRead;
            pCache->cacheErrWrite = errWrite;
            // pCache->firstDirtyByteOffset ... no init ...
            pCache->dirtyByteCount = 0;
            pCache->cacheSize = cacheSize;
            pCache->writeAccess = writeable;
            pCache->ftruncateAllowed = FALSE;
            pCache->callFnOnErr = callFnOnErr;
#if SENTINEL_SIZE

            memset(&(pCache->preSentinel), 0xE5, SENTINEL_SIZE);
            memset(&(pCache->postSentinel), 0xE5, SENTINEL_SIZE);

#endif

            CacheFill(pCache);

            if (pCache->cacheErrStatus) {
                fclose(pCache->pFile);
                CacheFree(&pCache);
            }
        }
    }

    return pCache;
} /* end of: CacheOpen */

TFileAddr CacheWhere(pCache) PCacheControl pCache; {
    SentinelTest(pCache);
    return ((pCache->nextRWAddr)-(pCache->curFileAddrOffset));
} /* end of: CacheWhere */

TFileAddr CacheSeek(pCache, fileAddr) PCacheControl pCache;
    TFileAddr fileAddr; {
    SentinelTest(pCache);

    if (fileAddr == CACHE_SEEK_EOF) {
        pCache->nextRWAddr = pCache->fileSize;
    } else {
        pCache->nextRWAddr = fileAddr + pCache->curFileAddrOffset;
    }

    return ((pCache->nextRWAddr)-(pCache->curFileAddrOffset));
} /* end of: CacheSeek */

TFileAddr CacheSeekRead(pCache, fileAddr, pData, size) PCacheControl pCache;
    TFileAddr fileAddr;
    void* pData;
    TULong size; {
    CacheSeek(pCache, fileAddr);
    CacheRead(pCache, pData, size);

    if (pCache->cacheErrStatus) {
        return CACHE_SEEK_EOF;
    }
    return CacheWhere(pCache);
} /* end of: CacheSeekRead */

TFileAddr CacheSeekWrite(pCache, fileAddr, pData, size) PCacheControl pCache;
    TFileAddr fileAddr;
    const void* pData;
    TULong size; {
    CacheSeek(pCache, fileAddr);
    CacheWrite(pCache, pData, size);

    if (pCache->cacheErrStatus) {
        return CACHE_SEEK_EOF;
    }
    return CacheWhere(pCache);
} /* end of: CacheSeekRead */

STATIC_FN(void) CacheClean(pCache)PCacheControl pCache; {
    /* Local variables */
    TULong bytesLeft;
    TFileAddr writeStartAddr;

    SentinelTest(pCache);

    if (pCache->dirtyByteCount) {
        writeStartAddr = pCache->cacheStartAddr + pCache->firstDirtyByteOffset;

        if (fseek(pCache->pFile, writeStartAddr, SEEK_SET)) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrSeek);
            }
            pCache->cacheErrStatus = pCache->cacheErrSeek;
            return;
        }
        bytesLeft = pCache->fileSize - writeStartAddr;

        if (pCache->dirtyByteCount > bytesLeft) {
            pCache->dirtyByteCount = (TUShort)bytesLeft;
        }

        if ((pCache->dirtyByteCount == 0) || (!fwrite(&(pCache->theCache[pCache->firstDirtyByteOffset]), pCache->dirtyByteCount, 1, pCache->pFile))) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrWrite);
            }
            pCache->cacheErrStatus = pCache->cacheErrWrite;
            return;
        }
        pCache->dirtyByteCount = 0;
    }
    SentinelTest(pCache);
} /* end of: CacheClean */

STATIC_FN(void) CacheFill(pCache)PCacheControl pCache; {
    /* Local variables */
    TULong bytesLeft;
    TUShort readAmount;

    CacheClean(pCache);

    if (pCache->cacheErrStatus) {
        return;
    }

    pCache->cacheStartAddr = pCache->nextRWAddr;

    if (pCache->cacheStartAddr < pCache->fileSize) {
        readAmount = pCache->cacheSize;
        bytesLeft = pCache->fileSize - pCache->cacheStartAddr;

        if (bytesLeft < readAmount) {
            readAmount = (TUShort)bytesLeft;
        }

        if (fseek(pCache->pFile, pCache->cacheStartAddr, SEEK_SET)) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrSeek);
            }
            pCache->cacheErrStatus = pCache->cacheErrSeek;
            return;
        }

        if ((readAmount == 0) || (!fread(pCache->theCache, readAmount, 1, pCache->pFile))) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrRead);
            }
            pCache->cacheErrStatus = pCache->cacheErrRead;
            return;
        }
        // WHY? pCache->theCache[readAmount] = '\0';
        pCache->cacheEndAddr = pCache->cacheStartAddr + readAmount - 1;
    } else if (pCache->fileSize) {
        pCache->cacheEndAddr = pCache->fileSize - 1;
    }

    SentinelTest(pCache);
} /* end of: CacheFill */

TULong CacheRead(pCache, pData, size) PCacheControl pCache;
    void* pData;
    TULong size; {
    /* Local variables */
    TFileAddr readEndAddr;
    TUShort cacheOffset;
    TULong numRead = 0;

    if (size > pCache->cacheSize) {
        /* Invalidate the cache and do a direct read */
        CacheClean(pCache);

        if (pCache->cacheErrStatus) {
            return 0;
        }
        pCache->cacheStartAddr = CACHE_START_DEACTIVATE;
        pCache->cacheEndAddr = 0;

        if (fseek(pCache->pFile, pCache->nextRWAddr, SEEK_SET)) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrSeek);
            }
            pCache->cacheErrStatus = pCache->cacheErrSeek;
            return 0;
        }
        numRead = fread(pData, 1, size, pCache->pFile);

        if (!numRead) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrRead);
            }
            pCache->cacheErrStatus = pCache->cacheErrRead;
            return 0;
        }
    } else {
        readEndAddr = pCache->nextRWAddr + size - 1;

        if ((pCache->nextRWAddr < pCache->cacheStartAddr) || (pCache->cacheEndAddr < readEndAddr)) {
            CacheFill(pCache); /* Need to refill the cache */

            if (pCache->cacheErrStatus) {
                return 0;
            }
        }
        /* Data is now in cache - copy it to the requestor */
        cacheOffset = (TUShort)(pCache->nextRWAddr - pCache->cacheStartAddr);
        memcpy(pData, &(pCache->theCache[cacheOffset]), size);
        numRead = size;
    }

    pCache->nextRWAddr += size;
    SentinelTest(pCache);
    return numRead;
} /* end of: CacheRead */


/*
 * This routine is similar to fgets.  However, it should be passed the
 *	maximum number of nonzero chars to read (rather than the max length).
 */
TBoolByte CacheReadString(pCache, pString, maxChars) PCacheControl pCache;
    char* pString;
    TUShort maxChars; {
    /* Local variables */
    TFileAddr readEndAddr;
    TUShort count = 0;
    char* pStr;

    SentinelTest(pCache);

    if (pCache->nextRWAddr < pCache->fileSize) {
        readEndAddr = pCache->nextRWAddr + maxChars;

        if (readEndAddr >= pCache->fileSize) {
            readEndAddr = pCache->fileSize - 1;
            maxChars = (TUShort)(readEndAddr - pCache->nextRWAddr + 1);
        }

        if ((pCache->nextRWAddr < pCache->cacheStartAddr) || (pCache->cacheEndAddr < readEndAddr)) {
            CacheFill(pCache);

            if (pCache->cacheErrStatus) {
                return FALSE;
            }
        }

        pStr = (char* )MEMCCPY(pString, &(pCache->theCache[(TUShort)(pCache->nextRWAddr - pCache->cacheStartAddr)]), '\n', maxChars);

        if (!pStr) {
            count = maxChars;
        } else {
            count = (TUShort)(pStr - pString);
#ifdef DOS

            if (count > 1) {
                if (pStr[-2] == '\r') {
                    pStr[-2] = '\n';
                    --count;
                    ++(pCache->nextRWAddr); /* Skips uncounted char */
                }
            }

#endif

        }

        pCache->nextRWAddr += count;
    }
#ifdef DOS

    if (*pString == '\032') count = 0;

#endif

    pString[count] = '\0';
    SentinelTest(pCache);

    return ((TBoolByte)(count != 0));
} /* end of: CacheReadString */


/*
 * CacheWriteString - This routine is similar to CacheWrite.  It also calls
 *	CacheWriteChar to output a newline.
 */
void CacheWriteString(pCache, pString, numChars) PCacheControl pCache;
    const char* pString;
    TUShort numChars; {
    /* Local variables */
    CacheWrite(pCache, pString, numChars);

    if (pCache->cacheErrStatus) {
        return;
    }
    CacheWriteChar(pCache, '\n');
} /* end of: CacheWriteString */


/*
 * CacheWriteChar - This routine outputs a character.  On DOS, it checks to see
 *	if the character is a newline, and treats it specially.
z */
void CacheWriteChar(pCache, c) PCacheControl pCache;
    char c; {
#ifdef DOS

    if (c == '\n') CacheWrite(pCache, "\r\n", 2); else

#endif

    CacheWrite(pCache, &c, 1);
} /* end of: CacheWriteChar */

TULong CacheWrite(pCache, pData, size) PCacheControl pCache;
    const void* pData;
    TULong size; {
    /* Local variables */
    TFileAddr writeEndAddr = pCache->nextRWAddr + size - 1;
    TUShort cacheOffset;
    TULong numWritten = 0;

    if (size > pCache->cacheSize) {
        /* Invalidate the cache and do a direct write */
        CacheClean(pCache);

        if (pCache->cacheErrStatus) {
            return 0;
        }

        pCache->cacheStartAddr = CACHE_START_DEACTIVATE;
        pCache->cacheEndAddr = 0;

        if (fseek(pCache->pFile, pCache->nextRWAddr, SEEK_SET)) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrSeek);
            }
            pCache->cacheErrStatus = pCache->cacheErrSeek;
            return 0;
        }
        numWritten = fwrite(pData, 1, size, pCache->pFile);

        if (!numWritten) {
            if (pCache->callFnOnErr) {
                CacheError(pCache->cacheErrWrite);
            }
            pCache->cacheErrStatus = pCache->cacheErrWrite;
            return 0;
        }
    } else {
        SentinelTest(pCache);

        if (writeEndAddr >= pCache->fileSize) {
            /* This write extends the file size */
            if ((pCache->cacheStartAddr <= pCache->nextRWAddr) && (writeEndAddr <= pCache->cacheStartAddr + pCache->cacheSize - 1)) {
                /* Extend the cache */
                pCache->cacheEndAddr = writeEndAddr;
            }
        }

        if ((pCache->nextRWAddr < pCache->cacheStartAddr) || (pCache->cacheEndAddr < writeEndAddr)) {
            CacheFill(pCache); /* Need to refill the cache */

            if (pCache->cacheErrStatus) {
                return 0;
            }

            if (writeEndAddr > pCache->cacheEndAddr) {
                pCache->cacheEndAddr = writeEndAddr; /* Extend the cache */
            }
        }

        /* Data area is now in cache - write the new data in */
        cacheOffset = (TUShort)(pCache->nextRWAddr - pCache->cacheStartAddr);
        memcpy(&(pCache->theCache[cacheOffset]), pData, size);
        numWritten = size;

        if (!(pCache->dirtyByteCount)) {
            pCache->firstDirtyByteOffset = cacheOffset;
            pCache->dirtyByteCount = (TUShort)size;
        } else {
            TUShort tempDirtyByteCount;

            if (cacheOffset < pCache->firstDirtyByteOffset) {
                pCache->dirtyByteCount += pCache->firstDirtyByteOffset - cacheOffset;
                pCache->firstDirtyByteOffset = cacheOffset;
            }
            tempDirtyByteCount = (TUShort)(cacheOffset + size - pCache->firstDirtyByteOffset);

            if (tempDirtyByteCount > pCache->dirtyByteCount) {
                pCache->dirtyByteCount = tempDirtyByteCount;
            }
        }
    }

    if (writeEndAddr >= pCache->fileSize) {
        pCache->fileSize = writeEndAddr + 1;
    }
    pCache->nextRWAddr += size;
    SentinelTest(pCache);
    return numWritten;
} /* end of: CacheWrite */

void CacheFileSizeSet(pCache, newFileSize) PCacheControl pCache;
    TULong newFileSize; {
    /* Local variables */
    int rc;

    if (pCache->pFile) {
        CacheClean(pCache);

        if (!(pCache->cacheErrStatus)) {
            pCache->fileSize = newFileSize;
            pCache->curFileAddrOffset = 0;
            pCache->nextRWAddr = 0;
            pCache->cacheStartAddr = CACHE_START_DEACTIVATE;
            pCache->cacheEndAddr = 0;
            // pCache->firstDirtyByteOffset ... no init ...
            pCache->dirtyByteCount = 0;

            if (pCache->ftruncateAllowed) {
                rc = ftruncate(FILENO(pCache->pFile), newFileSize);
            } else {
                rc = 0;
            }

            if (rc) {
                if (pCache->callFnOnErr) {
                    CacheError(pCache->cacheErrWrite);
                }
                pCache->cacheErrStatus = pCache->cacheErrWrite;
            } else {
                CacheFill(pCache);
            }
        }
    }
} /* end of: CacheFileSizeSet */

TBoolByte CacheClose(pCache, closeChar, closeErrCode) PCacheControl pCache;
    char closeChar;
    TErrCode closeErrCode; {
    /* Local variables */
    TBoolByte success = FALSE;

    if (pCache->pFile) {
        CacheClean(pCache);

        if (!(pCache->cacheErrStatus)) {
            success = TRUE;

            if (pCache->writeAccess) {
                if (fseek(pCache->pFile, 0, SEEK_END)) {
                    if (pCache->callFnOnErr) {
                        fclose(pCache->pFile);
                        pCache->pFile = NULL;
                        CacheError(pCache->cacheErrSeek);
                    }
                    pCache->cacheErrStatus = pCache->cacheErrSeek;
                    success = FALSE;
                } else if (fputc(closeChar, pCache->pFile) == EOF) {
                    success = FALSE;
                } else {
                    if (fclose(pCache->pFile) == EOF) {
                        success = FALSE;
                    }
                    pCache->pFile = NULL;
                }
            }
        }
    }

    if (pCache->pFile) {
        fclose(pCache->pFile);
        pCache->pFile = NULL;
    }

    if (!success) {
        if (pCache->callFnOnErr) {
            CacheFree(&pCache);
            CacheError(closeErrCode);
        }
    }

    SentinelTest(pCache);
    CacheFree(&pCache);

    return success;
} /* end of: CacheClose */

STATIC_FN(PCacheControl) CacheAlloc(cacheSize
#ifdef _DEBUG

, pFileName

#endif

)TUShort cacheSize;
#ifdef _DEBUG

    LPCSTR pFileName;

#endif

{
    /* Local variables */
    PCacheControl pCache = NULL;

    if (cacheSize <= MAX_CACHE_SIZE) {
        pCache = (PCacheControl)malloc(sizeof(TCacheControl));
    // !! In future, can improve memory usage if cache size < MAX
    }

#ifdef _DEBUG

    if (pCache) {
        FILE* logFile = fopen("DRF.LOG", F_ACCESS_APPEND_TEXT);

        if (logFile) {
            fprintf(logFile, "Alloc cache %lx for %s\n", (unsigned long)pCache, pFileName);
            fclose(logFile);
        }
    }

#endif

    return pCache;
} /* end of: CacheAlloc */

STATIC_FN(void) CacheFree(ppCache)PCacheControl* ppCache; {
    /* Local variables */
    PCacheControl pCache = *ppCache;

    if (pCache) {
#ifdef _DEBUG

        FILE* logFile = fopen("DRF.LOG", F_ACCESS_APPEND_TEXT);

        if (logFile) {
            fprintf(logFile, "Free cache %lx\n", (unsigned long)pCache);
            fclose(logFile);
        }

#endif

        free(pCache);
        * ppCache = NULL;
    }
} /* end of: CacheFree */
