/*
 * DRFBASE.C - This is the Resource File API file responsible for base API functions
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

static PLdrfFileInfo pValidFileInfoList;
static PLdrfTypeTree pValidTypeTreeList;

void LdrfInitialize(void) {

    pValidFileInfoList = NULL;
    pValidTypeTreeList = NULL;

    LdrfLangInitialize();
} /* end of: LdrfInitialize */

#ifdef DEFAULT_MEMORY_ALLOC_IMPLEMENTATION

void* LdrfAllocFn(TULong numBytes) {
    /* Local variables */
    void* p = NULL;

    if (numBytes) {
        p = (void* )malloc(numBytes);
    }

    return p;
} /* end of: LdrfAllocFn */

void LdrfFreeFn(void* * ppMem) {
    /* Local variables */
    void* pMem;

    if (ppMem) {
        pMem = *ppMem;

        if (pMem) {
            * ppMem = NULL;
            free(pMem);
        }
    }
} /* end of: LdrfFreeFn */

#endif

TLdrfErrCodes LdrfTransferStringOut(LPSTR pStrOut, PUShort pLen, LPCSTR pStr) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TUShort lenIn = (TUShort)strlen(pStr);

    if (pStrOut) {
        if (!pLen) {
            errCode = LDRF_ERR_PARAM;
        } else {
            TUShort lenOut = (TUShort)(*pLen - 1);

            if (lenIn <= lenOut) {
                strcpy(pStrOut, pStr);
                * pLen = lenIn + 1;
            } else {
                strncpy(pStrOut, pStr, lenOut);
                pStrOut[lenOut] = EOS;
                errCode = LDRF_ERR_TRUNC;
            }
        }
    }

    return errCode;
} /* end of: LdrfTransferStringOut */

TLdrfErrCodes LdrfNewFileInfo(PLdrfFileInfo* ppInfo, TLdrfFileType fileType) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;
    PLdrfFileInfo pInfo;

    pInfo = NEW(TLdrfFileInfo);

    if (pInfo) {
        errCode = LDRF_ERR_NONE;
        memset(pInfo, 0, offsetof(TLdrfFileInfo, u));
        pInfo->pNextValid = pValidFileInfoList;
        pValidFileInfoList = pInfo;
        pInfo->validator = VALIDATION_CODE_FILE_INFO;
        pInfo->fileType = fileType;

        switch (fileType) {
            case LDRF_CATALOG:
                memset(&(pInfo->u.cat), 0, sizeof(pInfo->u.cat));
                break;

            case LDRF_LANG_RESOURCE:
                memset(&(pInfo->u.lang), 0, sizeof(pInfo->u.lang));
                break;

            case LDRF_TYPE:
                memset(&(pInfo->u.type), 0, sizeof(pInfo->u.type));
                break;

            case LDRF_FPT:
                memset(&(pInfo->u.fpt), 0, sizeof(pInfo->u.fpt));
                break;

            default:
                errCode = LDRF_ERR_UNIMPL;
        }
    }

    * ppInfo = pInfo;

    return errCode;
} /* end of: LdrfNewFileInfo */

TLdrfErrCodes LdrfFindFileInfo(PLdrfFileInfo pInfo, PLdrfFileInfo* ppPrevInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_FILE_INFO;
    PLdrfFileInfo pPrevInfo = NULL;
    PLdrfFileInfo pScan = pValidFileInfoList;

    while ((pScan) && (pScan != pInfo)) {
        pPrevInfo = pScan;
        pScan = pScan->pNextValid;
    }

    if (pScan) {
        errCode = LDRF_ERR_NONE;

        if (ppPrevInfo) {
            * ppPrevInfo = pPrevInfo;
        }
    }

    return errCode;
} /* end of: LdrfFindFileInfo */

TLdrfErrCodes LdrfValidateFileInfoPointer(PLdrfFileInfo pInfo, TLdrfFileType fileType) {
    /* Local variables */
    PLdrfFileInfo pPrevInfo;
    TLdrfErrCodes errCode = LdrfFindFileInfo(pInfo, &pPrevInfo);

    if (!errCode) {
        if ((pInfo->validator != VALIDATION_CODE_FILE_INFO) || (!(pInfo->pCache))) {
            errCode = LDRF_ERR_FILE_INFO;
        } else if ((fileType != LDRF_WILD_CARD) && (pInfo->fileType != fileType)) {
            switch (fileType) {
                case LDRF_CATALOG:
                    errCode = LDRF_ERR_NOT_CATALOG;
                    break;

                case LDRF_LANG_RESOURCE:
                    errCode = LDRF_ERR_NOT_RESOURCE;
                    break;

                case LDRF_TYPE:
                    errCode = LDRF_ERR_NOT_TYPE;
                    break;

                case LDRF_FPT:
                    errCode = LDRF_ERR_NOT_FPT;
                    break;

                case LDRF_FORMAT:
                    errCode = LDRF_ERR_NOT_FORMAT;
                    break;

                default:
                    errCode = LDRF_ERR_INTERNAL;
            }
        }
    }

    if (!errCode) {
        /* Optimize future accesses - bring this one to the front of the list */
        if (pPrevInfo) {
            pPrevInfo->pNextValid = pInfo->pNextValid;
            pInfo->pNextValid = pValidFileInfoList;
            pValidFileInfoList = pInfo;
        }
    }

    return errCode;
} /* end of: LdrfValidateFileInfoPointer */

TLdrfErrCodes LdrfFreeFileInfo(PLdrfFileInfo* ppInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfFileInfo pInfo, pPrevInfo;

    if (ppInfo) {
        pInfo = *ppInfo;
        * ppInfo = NULL;

        if (!LdrfFindFileInfo(pInfo, &pPrevInfo)) {
            /* Remove this file info structure from the list */
            if (pPrevInfo) {
                pPrevInfo->pNextValid = pInfo->pNextValid;
            } else {
                pValidFileInfoList = pInfo->pNextValid;
            }
        }

        switch (pInfo->fileType) {
            case LDRF_CATALOG:
                errCode = LdrfFreeCatalogFileInfo(pInfo);
                break;

            case LDRF_LANG_RESOURCE:
                errCode = LdrfFreeLangFileInfo(pInfo);
                break;

            case LDRF_TYPE:
                errCode = LdrfFreeTypeFileInfo(pInfo);
                break;

            case LDRF_FPT:
                errCode = LdrfFreeFPTFileInfo(pInfo);
                break;

            default:
                errCode = LDRF_ERR_UNIMPL;
        }
        FREE(pInfo->pDesc);
        FREE(pInfo->pCreator);
        FREE(pInfo->pURL);
        FREE(pInfo);
    }

    return errCode;
} /* end of: LdrfFreeFileInfo */

TLdrfErrCodes LdrfFileSeek(PLdrfFileInfo pInfo, long seekSet) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;

    //if (!fseek(pInfo->pFile, seekSet, SEEK_SET)) {
    CacheSeek(pInfo->pCache, seekSet);

    if (!(pInfo->pCache->cacheErrStatus)) {
        errCode = LDRF_ERR_NONE;
    }

    return errCode;
} /* end of: LdrfFileSeek */

TLdrfErrCodes LdrfFileWhere(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;
    long retval;

    // retval = ftell(pInfo->pFile);
    retval = CacheWhere(pInfo->pCache);

    // if (retval != -1) {
    if (retval != (long)CACHE_SEEK_EOF) {
        pInfo->currentAddress = (TULong)retval;
        errCode = LDRF_ERR_NONE;
    }

    return errCode;
} /* end of: LdrfFileWhere */

TLdrfErrCodes LdrfFileReadBytes(PLdrfFileInfo pInfo, void* pBytes, TULong numBytes) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong count;

    if (!pBytes) {
        errCode = LDRF_ERR_PARAM;
    } else if (numBytes) {
        // count = fread(pBytes, 1, numBytes, pInfo->pFile);
        count = CacheRead(pInfo->pCache, pBytes, numBytes);

        if (count != numBytes) {
            errCode = LDRF_ERR_SYS;
        }
    }

    return errCode;
} /* end of: LdrfFileReadBytes */

TLdrfErrCodes LdrfFileReadString(PLdrfFileInfo pInfo, LPSTR pS, TUShort fixedLen) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;

    if (pS) {
        if (fixedLen) {
            errCode = LdrfFileReadBytes(pInfo, pS, fixedLen);

            if (!errCode) {
                pS[fixedLen - 1] = EOS;
            }
        } else {
            errCode = LDRF_ERR_UNIMPL;
        }
    }

    return errCode;
} /* end of: LdrfFileReadString */

TLdrfErrCodes LdrfFileReadLString(PLdrfFileInfo pInfo, LPSTR* ppS) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;
    TUByte len;

    if (ppS) {
        errCode = LdrfFileReadBytes(pInfo, &len, sizeof(len));

        if (!errCode) {
            * ppS = NEW_STRING(len);

            if (!(* ppS)) {
                errCode = LDRF_ERR_SYS;
            }
        }

        if ((!errCode) && (len)) {
            errCode = LdrfFileReadBytes(pInfo, *ppS, len);
        }

        if (!errCode) {
            (* ppS)[len] = EOS;
        }
    }

    return errCode;
} /* end of: LdrfFileReadLString */

TLdrfErrCodes LdrfFileReadLong(PLdrfFileInfo pInfo, TULong* pLong) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, pLong, sizeof(TULong));
#ifdef LCADRF_BYTE_SWAP

    TUByte* p = (TUByte* )pLong;
    TUByte swapTemp;

    if (!errCode) {
        swapTemp = p[0];
        p[0] = p[3];
        p[3] = swapTemp;
        swapTemp = p[1];
        p[1] = p[2];
        p[2] = swapTemp;
    }

#endif

    return errCode;
} /* end of: LdrfFileReadLong */

TLdrfErrCodes LdrfFileReadInt64 (PLdrfFileInfo pInfo, __DRF_PINT64 pInt64){
    /* Local variables */
    __DRF_INT64 temp;
    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, &temp, sizeof(temp));
    if (!errCode) {
        LdrfCopyAndSwapScalar(pInt64, &temp, sizeof(temp));
    }
    return errCode;
} /* end of: LdrfFileReadInt64 */


TLdrfErrCodes LdrfFileRead24(PLdrfFileInfo pInfo, TULong* pLong) {
    /* Local variables */
    TUByte* p = (TUByte* )pLong;
    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, p + 1, 3);

    if (!errCode) {
        p[0] = 0;
#ifdef LCADRF_BYTE_SWAP

        {
            TUByte swapTemp = p[0];
            p[0] = p[3];
            p[3] = swapTemp;
            swapTemp = p[1];
            p[1] = p[2];
            p[2] = swapTemp;
        }

#endif

    }

    return errCode;
} /* end of: LdrfFileRead24 */

void LdrfCopyAndSwapScalar(void *pTo, const void *pFrom, TULong numBytes)
{
#ifdef LCADRF_BYTE_SWAP
    TUByte* pToByte = ((TUByte *)pTo) + numBytes;
    const TUByte *pFromByte = (const TUByte *)pFrom;

    while (numBytes--)
        * (--pToByte) = *pFromByte++;
#else
    memcpy(pTo, pFrom, numBytes);
#endif
}

TLdrfErrCodes LdrfFileReadBigEndianBytes(PLdrfFileInfo pInfo, void* pBytes, TULong numBytes) {
#ifdef LCADRF_BYTE_SWAP
    /* Local variables */

    TUByte* pRead = LdrfAllocFn(numBytes);
    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, pRead, numBytes);

    if (errCode == LDRF_ERR_NONE) {
        LdrfCopyAndSwapScalar(pBytes, pRead, numBytes);
    }
    LdrfFreeFn(&pRead);

#else

    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, pBytes, numBytes);

#endif //    LCADRF_BYTE_SWAP

    return errCode;
}      /* end of: LdrfFileReadBigEndianBytes */

TLdrfErrCodes LdrfFileReadShort(PLdrfFileInfo pInfo, TUShort* pShort) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileReadBytes(pInfo, pShort, sizeof(TUShort));
#ifdef LCADRF_BYTE_SWAP

    TUByte* p = (TUByte* )pShort;

    if (!errCode) {
        TUByte swapTemp = p[0];
        p[0] = p[1];
        p[1] = swapTemp;
    }

#endif

    return errCode;
} /* end of: LdrfFileReadShort */

TLdrfErrCodes LdrfFileWriteBytes(PLdrfFileInfo pInfo, const void* pBytes, TULong numBytes) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_WRITE;
    TULong count;

    if (!pBytes) {
        errCode = LDRF_ERR_PARAM;
    } else {
        // count = fwrite(pBytes, 1, numBytes, pInfo->pFile);
        count = CacheWrite(pInfo->pCache, pBytes, numBytes);

        if (count == numBytes) {
            errCode = LDRF_ERR_NONE;
        }
    }

    return errCode;
} /* end of: LdrfFileWriteBytes */

TLdrfErrCodes LdrfFileWriteString(PLdrfFileInfo pInfo, LPCSTR pS, TUShort fixedLen) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;

    if (pS) {
        TUShort len = (TUShort)strlen(pS);

        if (fixedLen) {
            char* p = NEW_STRING(fixedLen - 1);

            if (!p) {
                errCode = LDRF_ERR_SYS;
            } else {
                if (len < fixedLen - 1) {
                    memset(p + len, ' ', fixedLen - len - 1);
                    strncpy(p, pS, len);
                    p[len] = ' ';
                } else {
                    strncpy(p, pS, fixedLen - 1);
                }
                p[fixedLen - 1] = EOS;
                errCode = LdrfFileWriteBytes(pInfo, p, fixedLen);
                FREE(p);
            }
        } else {
            errCode = LdrfFileWriteBytes(pInfo, pS, (TUShort)(len + 1));
        }
    }

    return errCode;
} /* end of: LdrfFileWriteString */

TLdrfErrCodes LdrfFileWriteLString(PLdrfFileInfo pInfo, LPCSTR pS, TUByte len) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;

    if (pS) {
        errCode = LdrfFileWriteBytes(pInfo, &len, sizeof(len));

        if ((!errCode) && (len)) {
            errCode = LdrfFileWriteBytes(pInfo, pS, len);
        }
    }

    return errCode;
} /* end of: LdrfFileWriteLString */

TLdrfErrCodes LdrfFileWriteLong(PLdrfFileInfo pInfo, const TULong* pLong) {
/* Local variables */
#ifdef LCADRF_BYTE_SWAP

    TUByte p[4];

    if (!pLong) {
        return LDRF_ERR_PARAM;
    } else {
        TUByte swapTemp;
        *((TULong* )p) = *pLong;
        //memcpy(p, pLong, sizeof(TULong));
        swapTemp = p[0];
        p[0] = p[3];
        p[3] = swapTemp;
        swapTemp = p[1];
        p[1] = p[2];
        p[2] = swapTemp;
    }
    return LdrfFileWriteBytes(pInfo, p, sizeof(TULong));

#else

    return LdrfFileWriteBytes(pInfo, pLong, sizeof(TULong));

#endif

} /* end of: LdrfFileWriteLong */

TLdrfErrCodes LdrfFileWriteInt64 (PLdrfFileInfo pInfo, const __DRF_PINT64 pInt64)
{
    /* Local variables */
    __DRF_INT64 temp;
    LdrfCopyAndSwapScalar(&temp, pInt64, sizeof(temp));
    return LdrfFileWriteBytes(pInfo, &temp, sizeof(temp));
}

TLdrfErrCodes LdrfFileWrite24(PLdrfFileInfo pInfo, const TULong* pLong) {
/* Local variables */
#ifdef LCADRF_BYTE_SWAP

    TUByte p[4];

    if (!pLong) {
        return LDRF_ERR_PARAM;
    } else {
        TUByte swapTemp;
        memcpy(p, pLong, 3);
        p[3] = p[0];
        swapTemp = p[1];
        p[1] = p[2];
        p[2] = swapTemp;
    }
    return LdrfFileWriteBytes(pInfo, p + 1, 3);

#else

    return LdrfFileWriteBytes(pInfo, ((TUByte* )pLong) + 1, 3);

#endif

} /* end of: LdrfFileWrite24 */

TLdrfErrCodes LdrfFileWriteShort(PLdrfFileInfo pInfo, const TUShort* pShort) {
/* Local variables */
#ifdef LCADRF_BYTE_SWAP

    TUByte p[2];

    if (!pShort) {
        return LDRF_ERR_PARAM;
    } else {
        TUByte swapTemp;
        *((TUShort* )p) = *pShort;
        //memcpy(p, pShort, sizeof(TUShort));
        swapTemp = p[0];
        p[0] = p[1];
        p[1] = swapTemp;
    }
    return LdrfFileWriteBytes(pInfo, p, sizeof(TUShort));

#else

    return LdrfFileWriteBytes(pInfo, pShort, sizeof(TUShort));

#endif

} /* end of: LdrfFileWriteShort */

TLdrfErrCodes LdrfFileWriteBigEndianBytes(PLdrfFileInfo pInfo, void* pBytes, TULong numBytes) {
#ifdef LCADRF_BYTE_SWAP
    /* Local variables */

    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TUByte* pWrite = LdrfAllocFn(numBytes);
    TUByte* pBigEndian = pWrite;
    TUByte* pLittleEndian = (TUByte* )pBytes + numBytes;
    TULong uCount = numBytes;

    while (uCount--)
        * pBigEndian++ = *(--pLittleEndian);

    errCode = LdrfFileWriteBytes(pInfo, pWrite, numBytes);
    LdrfFreeFn(&pWrite);

#else

    TLdrfErrCodes errCode = LdrfFileWriteBytes(pInfo, pBytes, numBytes);

#endif //    LCADRF_BYTE_SWAP

    return errCode;
}      /* end of: LdrfFileReadBigEndianBytes */

TLdrfErrCodes LdrfUpdateDataSize(PLdrfFileInfo pInfo) {
    TLdrfErrCodes errCode = LdrfFileWhere(pInfo);

    if (!errCode) {
        TULong newDataSize = pInfo->currentAddress - pInfo->dataAddress;

        if (newDataSize > pInfo->dataSize) {
            pInfo->dataSize = newDataSize;
            pInfo->updateHeader = T;
        }
    }

    pInfo->updateData = T;

    return errCode;
} /* end of: LdrfUpdateDataSize */

TLdrfErrCodes LdrfUpdateRecSize(PLdrfFileInfo pInfo, TULong startAddress, TUShort offsetRecSize) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong thisRecordSize = (pInfo->currentAddress - startAddress) - (offsetRecSize + 2);
    TUShort recSizeDummy = 0;

    if (thisRecordSize > 65535) {
        errCode = LDRF_ERR_INTERNAL;
    } else {
        recSizeDummy = (TUShort)thisRecordSize;
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, startAddress + offsetRecSize);
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &recSizeDummy);
    }

    return errCode;
} /* end of: LdrfUpdateRecSize */

TLdrfErrCodes LdrfMoveFileRec(PLdrfFileInfo pInfo, TULong oldAddress, TULong newAddress, TUShort offsetRecSize) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, oldAddress);
#define MAX_REC_SIZE_OFFSET 3

    TUByte buf[MAX_REC_SIZE_OFFSET];
    TUShort size = 0;

    if ((!errCode) && (offsetRecSize > MAX_REC_SIZE_OFFSET)) {
        errCode = LDRF_ERR_INTERNAL;
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, buf, offsetRecSize);
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &size);
    }

    if (!errCode) {
        PUByteArray pBA = NEW_ARRAY(TUByte, size);

        if (!pBA) {
            errCode = LDRF_ERR_SYS;
        } else {
            errCode = LdrfFileReadBytes(pInfo, pBA, size);

            if (!errCode) {
                errCode = LdrfFileSeek(pInfo, newAddress);
            }

            if (!errCode) {
                errCode = LdrfFileWriteBytes(pInfo, buf, offsetRecSize);
            }

            if (!errCode) {
                errCode = LdrfFileWriteShort(pInfo, &size);
            }

            if (!errCode) {
                errCode = LdrfFileWriteBytes(pInfo, pBA, size);
            }
            FREE(pBA);
        }
    }

    if (!errCode) {
        if (oldAddress == pInfo->dataAddress + SIZE_DATA_CRC) {
            pInfo->dataAddress = oldAddress + offsetRecSize + 2 + size - SIZE_DATA_CRC;
            pInfo->updateHeader = T;
        }
        errCode = LdrfUpdateDataSize(pInfo);
    }

    return errCode;
} /* end of: LdrfMoveFileRec */

TLdrfErrCodes LdrfReadDirectory(PLdrfFileInfo pInfo, TULong minimumSize, TULong numItems, PULong* ppDir) {
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PULong pDir = NULL;

    if ((numItems) || (minimumSize)) {
        if (minimumSize < numItems) {
            minimumSize = numItems;
        }

        if (!(* ppDir)) {
            pDir = NEW_ARRAY(TULong, minimumSize + 1);

            if (!pDir) {
                errCode = LDRF_ERR_SYS;
            }
            * ppDir = pDir;
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, pDir + 1, numItems * sizeof(TULong));
        }

        if (!errCode) {
#ifdef LCADRF_BYTE_SWAP

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
                /* By accident rather than design, in format-version "1" of
                 * the CAT, <lang>, and FPT files, and format-version "2" of
                 * the TYP file, the directory entries weren't byte swapped.
                 * This is corrected below for the later versions ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#else

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer < 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer < 2))) {
                /* The earlier versions, as noted above, were byte-swapped
                 * when they weren't supposed to be ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#endif

        }
    }

    return errCode;
} /* end of: LdrfReadDirectory */

TLdrfErrCodes LdrfWriteDirectory(PLdrfFileInfo pInfo, TULong numItems, PULong pDir) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    if (numItems) {
        if (!pDir) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
#ifdef LCADRF_BYTE_SWAP
            /* If the info in the directory must be swapped before writing,
             * it must then be swapped again after (to put it back in machine-
             * usable form). */

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
                /* By accident rather than design, in format-version "1" of
                 * the CAT, <lang>, and FPT files, and format-version "2" of
                 * the TYP file, the directory entries weren't byte swapped.
                 * This is corrected below for the later versions ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#else

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer < 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer < 2))) {
                /* The earlier versions, as noted above, were byte-swapped
                 * when they weren't supposed to be ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#endif

            errCode = LdrfFileWriteBytes(pInfo, pDir + 1, numItems * sizeof(TULong));
#ifdef LCADRF_BYTE_SWAP

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
                /* By accident rather than design, in format-version "1" of
                 * the CAT, <lang>, and FPT files, and format-version "2" of
                 * the TYP file, the directory entries weren't byte swapped.
                 * This is corrected below for the later versions ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#else

            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer < 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer < 2))) {
                /* The earlier versions, as noted above, were byte-swapped
                 * when they weren't supposed to be ... */
                unsigned int i;
                TUByte swapTemp;
                TUByte* p;

                for ( i = 1; i <= numItems; ++i ) {
                    p = (TUByte* )(pDir + i);
                    swapTemp = p[0];
                    p[0] = p[3];
                    p[3] = swapTemp;
                    swapTemp = p[1];
                    p[1] = p[2];
                    p[2] = swapTemp;
                }
            }

#endif

        }
    }

    return errCode;
} /* end of: LdrfWriteDirectory */

TLdrfErrCodes LdrfFindInUnsortedDir(PLdrfFileInfo pInfo, TULong* pDir, TULong numEntries, LPCSTR pName, PULong pEntry, PLdrfCompareFn fn) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong entry = 1;

    if ((!pDir) || (!fn)) {
        errCode = LDRF_ERR_INTERNAL;
    }

    while ((!errCode) && (entry <= numEntries)) {
        int rc;
        errCode = fn(pInfo, pName, pDir[entry], &rc);

        if (!errCode) {
            if (rc == 0) {
                break;
            } else {
                ++entry;
            }
        }
    }

    if ((!errCode) && (entry > numEntries)) {
        errCode = LDRF_ERR_NOT_FOUND;
    }

    if ((!errCode) && (pEntry)) {
        * pEntry = entry;
    }

    return errCode;
} /* end of: LdrfFindInUnsortedDir */

TLdrfErrCodes LdrfFindInSortedDir(PLdrfFileInfo pInfo, TULong* pDir, TULong numEntries, LPCSTR pName, PULong pEntry, PLdrfCompareFn fn) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong entry = 1;

    if (!numEntries) {
        errCode = LDRF_ERR_NOT_FOUND;
    } else if ((!pDir) || (!fn)) {
        errCode = LDRF_ERR_INTERNAL;
    }

    while ((!errCode) && (entry <= numEntries)) {
        int rc;
        errCode = fn(pInfo, pName, pDir[entry], &rc);

        if (!errCode) {
            if (rc <= 0) {
                if (rc != 0) {
                    errCode = LDRF_ERR_NOT_FOUND;
                }
                break;
            } else {
                ++entry;
            }
        }
    }

    if ((!errCode) && (entry > numEntries)) {
        errCode = LDRF_ERR_NOT_FOUND;
    }

    if ((!errCode) && (pEntry)) {
        * pEntry = entry;
    }

    return errCode;
} /* end of: LdrfFindInSortedDir */

TLdrfErrCodes LdrfSearchDirForMinEntry(PULong pDir, TULong numEntries, PULong pMinEntry, PULong pMinValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong index, minEntry = 0;
    TULong minValue = *pMinValue;

    for ( index = 1; index <= numEntries; ++index ) {
        if ((pDir[index]) && (pDir[index] <= minValue)) {
            if (pDir[index] < minValue) {
                minEntry = index;
                minValue = pDir[index];
            } else if (minEntry) {
                errCode = LDRF_ERR_INTERNAL;
                break;
            } else {
                minEntry = index;
            }
        }
    }
    * pMinEntry = minEntry;
    * pMinValue = minValue;

    return errCode;
} /* end of: LdrfSearchDirForMinEntry */

TLdrfErrCodes LdrfIncreaseCapacity(PULong pCapacity, TULong capacityMax, PULong pHdrSizeIncrement, PULong* ppDir1, PULong* ppDir2, PULong* ppDir3) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong hdrSizeIncrement = 0;
    TULong oldCapacity = *pCapacity, newCapacity = oldCapacity;
    TUShort numDirs = 0;

    if (ppDir1) ++numDirs;

    if (ppDir2) ++numDirs;

    if (ppDir3) ++numDirs;

    if (!numDirs) {
        errCode = LDRF_ERR_INTERNAL;
    } else if (oldCapacity >= capacityMax) {
        errCode = LDRF_ERR_FULL;
    } else if (oldCapacity >= capacityMax - CAPACITY_INCREMENT) {
        newCapacity = capacityMax;
        hdrSizeIncrement = (capacityMax - oldCapacity) * (numDirs * sizeof(TULong));
    } else {
        newCapacity = oldCapacity + CAPACITY_INCREMENT;
        hdrSizeIncrement = CAPACITY_INCREMENT * (numDirs * sizeof(TULong));

        if (!oldCapacity) {
            hdrSizeIncrement += (numDirs * sizeof(TULong)); /* Burn the 0th entries */
        }
    }

    if ((!errCode) && (ppDir1)) {
        errCode = LdrfIncreaseDirectory(ppDir1, oldCapacity, newCapacity);
    }

    if ((!errCode) && (ppDir2)) {
        errCode = LdrfIncreaseDirectory(ppDir2, oldCapacity, newCapacity);
    }

    if ((!errCode) && (ppDir3)) {
        errCode = LdrfIncreaseDirectory(ppDir3, oldCapacity, newCapacity);
    }

    * pCapacity = newCapacity;
    * pHdrSizeIncrement = hdrSizeIncrement;

    return errCode;
} /* end of: LdrfIncreaseCapacity */

TLdrfErrCodes LdrfIncreaseDirectory(PULong* ppDir, TULong oldCapacity, TULong newCapacity) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;
    PULong pOldDir = *ppDir;
    PULong pNewDir = NEW_ARRAY(TULong, newCapacity + 1);

    if (pNewDir) {
        errCode = LDRF_ERR_NONE;
        * ppDir = pNewDir;

        if (pOldDir) {
            memcpy(pNewDir, pOldDir, (oldCapacity + 1) * sizeof(TULong));
            FREE(pOldDir);
        }
    }

    return errCode;
} /* end of: LdrfIncreaseDirectory */

TLdrfErrCodes LdrfNewTypeTree(PLdrfTypeTree* ppTypeTree) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfTypeTree pTypeTree;

    pTypeTree = NEW(TLdrfTypeTree);

    if (!pTypeTree) {
        errCode = LDRF_ERR_SYS;
    } else {
        memset(pTypeTree, 0, sizeof(TLdrfTypeTree));
        pTypeTree->validator = VALIDATION_CODE_TYPE_TREE;
        pTypeTree->pNextValid = pValidTypeTreeList;
        pValidTypeTreeList = pTypeTree;
        pTypeTree->state = NVT_STATE_NEW;
    }

    * ppTypeTree = pTypeTree;

    return errCode;
} /* end of: LdrfNewTypeTree */

TLdrfErrCodes LdrfFindTypeTree(PLdrfTypeTree pTypeTree, PLdrfTypeTree* ppPrevTree) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_TYPE_TREE;
    PLdrfTypeTree pPrevTree = NULL;
    PLdrfTypeTree pScan = pValidTypeTreeList;

    while ((pScan) && (pScan != pTypeTree)) {
        pPrevTree = pScan;
        pScan = pScan->pNextValid;
    }

    if (pScan) {
        errCode = LDRF_ERR_NONE;

        if (ppPrevTree) {
            * ppPrevTree = pPrevTree;
        }
    }

    return errCode;
} /* end of: LdrfFindTypeTree */

TLdrfErrCodes LdrfValidateTypeTreePointer(PLdrfTypeTree pTypeTree) {
    /* Local variables */
    PLdrfTypeTree pPrevTree;
    TLdrfErrCodes errCode = LdrfFindTypeTree(pTypeTree, &pPrevTree);

    if ((!errCode) && (pTypeTree->validator != VALIDATION_CODE_TYPE_TREE)) {
        errCode = LDRF_ERR_TYPE_TREE;
    }

    if (!errCode) {
        /* Optimize future accesses - bring this one to the front of the list */
        if (pPrevTree) {
            pPrevTree->pNextValid = pTypeTree->pNextValid;
            pTypeTree->pNextValid = pValidTypeTreeList;
            pValidTypeTreeList = pTypeTree;
        }
    }

    return errCode;
} /* end of: LdrfValidateTypeTreePointer */

void LdrfFreeTypeTreePtr(PLdrfTypeTree* ppTypeTree) {
    /* Local variables */
    PLdrfTypeTree pTypeTree, pPrevTree;

    if (ppTypeTree) {
        pTypeTree = *ppTypeTree;
        * ppTypeTree = NULL;

        if (LdrfFindTypeTree(pTypeTree, &pPrevTree) == LDRF_ERR_NONE) {
            /* Remove this type tree structure from the list */
            if (pPrevTree != NULL) {
                pPrevTree->pNextValid = pTypeTree->pNextValid;
            } else {
                pValidTypeTreeList = pTypeTree->pNextValid;
            }
        }
        FREE(pTypeTree);
    }
} /* end of: LdrfFreeTypeTreePtr */

void CacheError(TErrCode errCode) {
} /* end of: CacheError */

TLdrfErrCodes LdrfReadRegistryKeyString(HKEY rootKey, LPCTSTR subKey, LPCTSTR valueName, TCHAR* buf, DWORD* pBufLen) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    HKEY hKey;

    if (RegOpenKeyEx(rootKey, subKey, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS) {
        errCode = LDRF_ERR_SYS;
    }

    if (!errCode) {
        if ((buf) && (pBufLen)) {
            long res = RegQueryValueEx(hKey, valueName, NULL, NULL, (LPBYTE)buf, pBufLen);

            if (res != ERROR_SUCCESS) {
                errCode = LDRF_ERR_SYS;
            }
        }
        RegCloseKey(hKey);
    }

    return errCode;
} /* end of: LdrfReadRegistryKeyString */

void LdrfGetScalar64Min(TBool isSigned, __DRF_PINT64 pMinValue)
{
    __drf_uint64 minValue;
    memset(&minValue, 0, sizeof(minValue));
    if (isSigned)
    {
        // Set sign bit
        minValue.value[0] |= 0x80;
    }
    LdrfCopyAndSwapScalar(pMinValue, &minValue, sizeof(*pMinValue));
}

void LdrfGetScalar64Max(TBool isSigned, __DRF_PINT64 pMaxValue)
{
    __drf_uint64 maxValue;
    memset(&maxValue, 0xff, sizeof(maxValue));
    if (isSigned)
    {
        // Clear sign bit
        maxValue.value[0] &= 0x7f;
    }
    LdrfCopyAndSwapScalar(pMaxValue, &maxValue, sizeof(*pMaxValue));
}

void LdrfGetScalar64Zero(__DRF_PINT64 pValue)
{
    memset(pValue, 0, sizeof(*pValue));
}

void LdrfGetScalar64NegativeOne(__DRF_PINT64 pValue)
{
    memset(pValue, 0xff, sizeof(*pValue));
}