/*
 * DRFGEN.C - This is the Resource File API file responsible for general file access
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"
#include "version.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

TLdrfErrCodes LdrfGetFileType(LPCSTR filePath, TLdrfFileType* pFileType) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TLdrfFileType fileType = LDRF_WILD_CARD;
    FILE* pFile = fopen(filePath, F_ACCESS_READ_BINARY);

    if (!pFile) {
        errCode = LDRF_ERR_NOT_FOUND;
    } else {
        char buf[4];
        errCode = LDRF_ERR_FILE_TYPE;

        if ((fread(buf, 1, 4, pFile) == 4) && ((buf[0] == 'S') || (buf[0] == 'U'))) {
            errCode = LDRF_ERR_NONE;

            if ((buf[1] == 'R') && (buf[2] == 'E') && (buf[3] == 'S')) {
                fileType = LDRF_LANG_RESOURCE;
            } else if ((buf[1] == 'T') && (buf[2] == 'Y') && (buf[3] == 'P')) {
                fileType = LDRF_TYPE;
            } else if ((buf[1] == 'F') && (buf[2] == 'P') && (buf[3] == 'T')) {
                fileType = LDRF_FPT;
            } else {
                errCode = LDRF_ERR_FILE_TYPE;
            }
        }
        fclose(pFile);
    }

    if (pFileType) {
        * pFileType = fileType;
    }

    return errCode;
} /* end of: LdrfGetFileType */

LDRF_FN LdrfCheckHeaderCRC(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);
    TUShort newCRC;

    if (!errCode) {
        errCode = LdrfComputeHeaderCRC(pInfo, &newCRC);
    }

    if ((!errCode) && (newCRC != pInfo->headerCRC)) {
        errCode = LDRF_ERR_CRC;
    }

    return errCode;
} /* end of: LdrfCheckHeaderCRC */

extern TLdrfErrCodes LdrfComputeHeaderCRC(PLdrfFileInfo pInfo, PUShort pCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileWhere(pInfo);
    TULong rememberWhere = pInfo->currentAddress;
    TULong sizeFirstPartHeader = SIZE_FIRST_PART_HDR_S;

    if (!errCode) {
        if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
            sizeFirstPartHeader = SIZE_FIRST_PART_HDR_L;
        }
        * pCRC = 0;
        errCode = LdrfComputeFileCRC(pInfo, 0, sizeFirstPartHeader, pCRC);
    }

    if (!errCode) {
        TULong secondPartStartAddr = sizeFirstPartHeader + SIZE_HDR_CRC;
        TULong secondPartSize = pInfo->headerSize - (sizeFirstPartHeader + SIZE_HDR_CRC);
        errCode = LdrfComputeFileCRC(pInfo, secondPartStartAddr, secondPartSize, pCRC);
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, rememberWhere);
    }

    return errCode;
} /* end of: LdrfComputeHeaderCRC */

LDRF_FN LdrfCheckDataCRC(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);
    TUShort newCRC, currentCRC;

    if ((!errCode) && ((pInfo->dataAddress) && (pInfo->dataSize > SIZE_DATA_CRC))) {
        TULong rememberWhere;
        errCode = LdrfFileWhere(pInfo);

        if (!errCode) {
            rememberWhere = pInfo->currentAddress;
            errCode = LdrfFileSeek(pInfo, pInfo->dataAddress);
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &currentCRC);
        }

        if (!errCode) {
            errCode = LdrfComputeDataCRC(pInfo, &newCRC);
        }

        if (!errCode) {
            errCode = LdrfFileSeek(pInfo, rememberWhere);
        }

        if ((!errCode) && (newCRC != currentCRC)) {
            errCode = LDRF_ERR_CRC;
        }
    }

    return errCode;
} /* end of: LdrfCheckDataCRC */

extern TLdrfErrCodes LdrfComputeDataCRC(PLdrfFileInfo pInfo, PUShort pCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileWhere(pInfo);
    TULong rememberWhere = pInfo->currentAddress;

    if (!errCode) {
        * pCRC = 0;
        errCode = LdrfComputeFileCRC(pInfo, pInfo->dataAddress + 2, pInfo->dataSize - 2, pCRC);
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, rememberWhere);
    }

    return errCode;
} /* end of: LdrfComputeHeaderCRC */

LDRF_FN LdrfCheckCRC(PUByteArray pBlock, TULong size, TUShort oldCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    if (oldCRC != LdrfComputeCRC(pBlock, size, 0)) {
        errCode = LDRF_ERR_CRC;
    }

    return errCode;
} /* end of: LdrfCheckCRC */

TLdrfErrCodes LdrfComputeFileCRC(PLdrfFileInfo pInfo, TULong startAddr, TULong size, PUShort pCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, startAddr);
    TUShort blockSize, crc = *pCRC;
#define MAX_FILE_BLOCK_SIZE 2048

    TUByte block[MAX_FILE_BLOCK_SIZE];

    while ((!errCode) && (size)) {
        if (size > MAX_FILE_BLOCK_SIZE) {
            blockSize = MAX_FILE_BLOCK_SIZE;
        } else {
            blockSize = (TUShort)size;
        }
        errCode = LdrfFileReadBytes(pInfo, block, blockSize);

        if (!errCode) {
            crc = LdrfComputeCRC(block, blockSize, crc);
            size -= blockSize;
        }
    }

    if (!errCode) {
        * pCRC = crc;
    }

    return errCode;
} /* end of: LdrfComputeFileCRC */

TUShort LdrfComputeCRC(PUByteArray pBlock, TULong size, TUShort startCRC) {
    /* Local variables */
    static TUShort crcTable[16] =
        {
        0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401, 0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400
        };

    TULong count;
    TUShort crc = startCRC, r, value;

    /* For each byte in the block, the routine computes the CRC iteratively.  The
     * start CRC should begin at a known value, such as 0.  For chaining CRC calculations
     * over multiple blocks (such as disk blocks, etc) the subsequent calls for the all except
     * the first block must pass in the result from the previous block as the start for the
     * next block.  The value returned for the last block is thus the composite CRC for all
     * the blocks as taken together.  CRC calculations are order-dependent, thus the results
     * may be different if the order is changed. */

    for ( count = 0; count < size; ++count ) {
        value = pBlock[count];
        r = crcTable[crc & 0xF] ^ crcTable[value & 0xF];
        crc = (crc >> 4) & 0xFFF;
        crc ^= r;
        r = crcTable[crc & 0xF] ^ crcTable[(value >> 4) & 0xF];
        crc = (crc >> 4) & 0xFFF;
        crc ^= r;
    }

    return crc;
} /* end of: LdrfComputeCRC */

TLdrfErrCodes LdrfUpdateHeaderCRC(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;

    errCode = LdrfFileWhere(pInfo);

    if (!errCode) {
        pInfo->headerSize = pInfo->currentAddress;

        if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
            errCode = LdrfFileSeek(pInfo, (SIZE_FIRST_PART_HDR_L - SIZE_HDR_SIZE_L));

            if (!errCode) {
                errCode = LdrfFileWriteLong(pInfo, &(pInfo->headerSize));
            }
        } else {
            /* In version 1 of LANG, FPT, and CAT files, and version 2 of
             * TYP files, the header size was mistakenly written as a short
             * instead of a long.  Part of the reason for this is that the
             * function prototypes used "void *" instead of "long *" as
             * should have been.  Those prototypes have now been fixed, and
             * later format-versions of the files use the code above, instead
             * of the wrong code below.  The net effect of the writing of
             * the smaller header size is that a resource file cannot grow
             * to "very large".  This is characterized elsewhere,
             * as part of the EPR for the problem. This code, still needed
             * to support the older format files, has been modified for
             * non-PC platform portability. */
            errCode = LdrfFileSeek(pInfo, (SIZE_FIRST_PART_HDR_S - SIZE_HDR_SIZE_S));

            if (!errCode) {
                TUShort temp = (TUShort)(pInfo->headerSize);

                if ((TULong)temp != pInfo->headerSize) {
                    errCode = LDRF_ERR_INTERNAL;
                } else {
                    errCode = LdrfFileWriteShort(pInfo, &temp);
                }
            }
        }
    }

    if (!errCode) {
        errCode = LdrfComputeHeaderCRC(pInfo, &(pInfo->headerCRC));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->headerCRC));
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, pInfo->headerSize);
    }

    return errCode;
} /* end of: LdrfUpdateHeaderCRC */

TLdrfErrCodes LdrfUpdateDataCRC(PLdrfFileInfo pInfo) {
    /* Local variables */
    TUShort dataCRC;
    TULong rememberWhere;
    TLdrfErrCodes errCode = LdrfFileWhere(pInfo);

    if ((!errCode) && ((!(pInfo->dataAddress)) || (!(pInfo->dataSize)))) {
        errCode = LDRF_ERR_INTERNAL;
    }

    if (!errCode) {
        rememberWhere = pInfo->currentAddress;
        errCode = LdrfComputeDataCRC(pInfo, &dataCRC);
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, pInfo->dataAddress);
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &dataCRC);
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, rememberWhere);
    }

    return errCode;
} /* end of: LdrfUpdateDataCRC */

TLdrfErrCodes LdrfReadFileHeader(PLdrfFileInfo pInfo, TBool checkCRC) {
    /* Local variables */
    TLdrfErrCodes errCode;
    char buf[4];

    switch (pInfo->fileType) {
        case LDRF_CATALOG:
            errCode = LDRF_ERR_NOT_CATALOG;
            break;

        case LDRF_LANG_RESOURCE:
            errCode = LDRF_ERR_NOT_RESOURCE;
            break;

        case LDRF_TYPE:
            errCode = LDRF_ERR_NOT_TYPE;
            break;

        case LDRF_FPT:
            errCode = LDRF_ERR_NOT_FPT;
            break;

        default:
            errCode = LDRF_ERR_FILE_TYPE;
            break;
    }

    if (errCode != LDRF_ERR_FILE_TYPE) {
        TLdrfErrCodes ec2 = errCode;
        TUByte majFmtLow, majFmtHigh;

        errCode = LdrfSupportedFormats(pInfo->fileType, &majFmtLow, &majFmtHigh);

        if (!errCode) {
            errCode = LdrfFileSeek(pInfo, 0L);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, buf, SIZE_HDR_PREFIX);
        }

        if (!errCode) {
            switch (pInfo->fileType) {
                case LDRF_CATALOG:
                    if (memcmp(buf, "SCAT", 4)) {
                        errCode = ec2;
                    }
                    break;

                case LDRF_LANG_RESOURCE:
                    if (buf[0] == 'S') {
                        pInfo->userFlag = F;
                    } else if (buf[0] == 'U') {
                        pInfo->userFlag = T;
                    } else {
                        errCode = ec2;
                    }

                    if ((!errCode) && (memcmp(&buf[1], "RES", 3))) {
                        errCode = ec2;
                    }
                    break;

                case LDRF_TYPE:
                    if (buf[0] == 'S') {
                        pInfo->userFlag = F;
                    } else if (buf[0] == 'U') {
                        pInfo->userFlag = T;
                    } else {
                        errCode = ec2;
                    }

                    if ((!errCode) && (memcmp(&buf[1], "TYP", 3))) {
                        errCode = ec2;
                    }
                    break;

                case LDRF_FPT:
                    if (buf[0] == 'S') {
                        pInfo->userFlag = F;
                    } else if (buf[0] == 'U') {
                        pInfo->userFlag = T;
                    } else {
                        errCode = ec2;
                    }

                    if ((!errCode) && (memcmp(&buf[1], "FPT", 3))) {
                        errCode = ec2;
                    }
                    break;

                default:
                    errCode = LDRF_ERR_UNIMPL;
            }
        }

        if (!errCode) {
            if (!(pInfo->pDesc)) {
                pInfo->pDesc = NEW_STRING(SIZE_DESC_STRING);
            }

            if (!(pInfo->pCreator)) {
                pInfo->pCreator = NEW_STRING(SIZE_CRE_STRING);
            }

            if (!(pInfo->pURL)) {
                pInfo->pURL = NEW_STRING(SIZE_URL_STRING);
            }

            if ((!(pInfo->pDesc)) || (!(pInfo->pCreator)) || (!(pInfo->pURL))) {
                errCode = LDRF_ERR_SYS;
            }
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, pInfo->pDesc, SIZE_DESC_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, pInfo->pCreator, SIZE_CRE_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, pInfo->pURL, SIZE_URL_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->resDescScope), sizeof(pInfo->resDescScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &(pInfo->resDescIndex));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->resCreScope), sizeof(pInfo->resCreScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &(pInfo->resCreIndex));
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->majorDataVer));
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->year));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->month), sizeof(pInfo->month));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->day), sizeof(pInfo->day));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->hour), sizeof(pInfo->hour));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->minute), sizeof(pInfo->minute));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->second), sizeof(pInfo->second));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->minorDataVer), sizeof(pInfo->minorDataVer));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->majorFmtVer), sizeof(pInfo->majorFmtVer));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &(pInfo->minorFmtVer), sizeof(pInfo->minorFmtVer));
        }

        if ((!errCode) && ((pInfo->majorFmtVer < majFmtLow) || (pInfo->majorFmtVer > majFmtHigh))) {
            errCode = LDRF_ERR_FMT_VERSION;
        }

        if (!errCode) {
            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
                errCode = LdrfFileReadLong(pInfo, &(pInfo->headerSize));
            } else {
                /* In version 1 of LANG, FPT, and CAT files, and version 2 of
                 * TYP files, the header size was mistakenly written as a short
                 * instead of a long.  Part of the reason for this is that the
                 * function prototypes used "void *" instead of "long *" as
                 * should have been.  Those prototypes have now been fixed,
                 * and later format-versions of the files use the code above,
                 * instead of the wrong code below (which has also been made
                 * platform-independent).  The net effect of the writing of
                 * the smaller header size is that a resource file cannot
                 * grow to "very large".  This is characterized elsewhere
                 * as part of the EPR for the problem. */
                TUShort temp;
                errCode = LdrfFileReadShort(pInfo, &temp);
                pInfo->headerSize = temp;
            }
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->headerCRC));
        }

        if (!errCode) {
            errCode = LdrfFileReadLong(pInfo, &(pInfo->dataAddress));
        }

        if (!errCode) {
            errCode = LdrfFileReadLong(pInfo, &(pInfo->dataSize));
        }

        if ((!errCode) && (checkCRC)) {
            errCode = LdrfCheckHeaderCRC(pInfo);

            if (!errCode) {
                errCode = LdrfCheckDataCRC(pInfo);
            }
        }
    }

    return errCode;
} /* end of: LdrfReadFileHeader */

TLdrfErrCodes LdrfInitFileHeader(PLdrfFileInfo pInfo, TUByte majFmt) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    pInfo->pDesc = NEW_STRING(MAX_FILE_INFO_STRING - 4);
    pInfo->pCreator = NEW_STRING(MAX_FILE_INFO_STRING);
    pInfo->pURL = NEW_STRING(MAX_FILE_INFO_STRING);

    if ((!(pInfo->pDesc)) || (!(pInfo->pCreator)) || (!(pInfo->pURL))) {
        errCode = LDRF_ERR_SYS;
    } else {
        pInfo->userFlag = T;
        pInfo->majorFmtVer = majFmt;
        memset(pInfo->pDesc, ' ', MAX_FILE_INFO_STRING - 4);
        pInfo->pDesc[MAX_FILE_INFO_STRING - 4] = EOS;
        memset(pInfo->pCreator, ' ', MAX_FILE_INFO_STRING);
        pInfo->pCreator[MAX_FILE_INFO_STRING] = EOS;
        memset(pInfo->pURL, ' ', MAX_FILE_INFO_STRING);
        pInfo->pURL[MAX_FILE_INFO_STRING] = EOS;

        if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
            pInfo->headerSize = SIZE_FIXED_HDR_L;
        } else {
            pInfo->headerSize = SIZE_FIXED_HDR_S;
        }

        pInfo->needsHdrInfo = T;
        pInfo->needsVerInfo = T;
        pInfo->updateHeader = T;
    }

    return errCode;
} /* end of: LdrfInitFileHeader */

TLdrfErrCodes LdrfWriteFileHeader(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    char buf[4];

    if ((!errCode) && ((pInfo->needsHdrInfo) || (pInfo->needsVerInfo))) {
        errCode = LDRF_ERR_INCOMPLETE;
    }

    if (!errCode) {
        errCode = LdrfFileSeek(pInfo, 0L);

        if (!errCode) {
            if (pInfo->userFlag) {
                buf[0] = 'U';
            } else {
                buf[0] = 'S';
            }

            switch (pInfo->fileType) {
                case LDRF_CATALOG:
                    memcpy(&(buf[1]), "CAT", 3);
                    break;
#ifdef LCADRF_WRITE

                case LDRF_LANG_RESOURCE:
                    memcpy(&(buf[1]), "RES", 3);
                    break;

                case LDRF_TYPE:
                    memcpy(&(buf[1]), "TYP", 3);
                    break;

                case LDRF_FPT:
                    memcpy(&(buf[1]), "FPT", 3);
                    break;

#endif

                default:
                    errCode = LDRF_ERR_INTERNAL;
            }
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, buf, SIZE_HDR_PREFIX);
        }

        if (!errCode) {
            errCode = LdrfFileWriteString(pInfo, pInfo->pDesc, SIZE_DESC_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileWriteString(pInfo, pInfo->pCreator, SIZE_CRE_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileWriteString(pInfo, pInfo->pURL, SIZE_URL_STRING);
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->resDescScope), sizeof(pInfo->resDescScope));
        }

        if (!errCode) {
            errCode = LdrfFileWrite24(pInfo, &(pInfo->resDescIndex));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->resCreScope), sizeof(pInfo->resCreScope));
        }

        if (!errCode) {
            errCode = LdrfFileWrite24(pInfo, &(pInfo->resCreIndex));
        }

        if (!errCode) {
            errCode = LdrfFileWriteShort(pInfo, &(pInfo->majorDataVer));
        }

        if (!errCode) {
            errCode = LdrfFileWriteShort(pInfo, &(pInfo->year));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->month), sizeof(pInfo->month));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->day), sizeof(pInfo->day));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->hour), sizeof(pInfo->hour));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->minute), sizeof(pInfo->minute));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->second), sizeof(pInfo->second));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->minorDataVer), sizeof(pInfo->minorDataVer));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->majorFmtVer), sizeof(pInfo->majorFmtVer));
        }

        if (!errCode) {
            errCode = LdrfFileWriteBytes(pInfo, &(pInfo->minorFmtVer), sizeof(pInfo->minorFmtVer));
        }

        if (!errCode) {
            if (((pInfo->fileType == LDRF_TYPE) && (pInfo->majorFmtVer >= 3)) || ((pInfo->fileType != LDRF_TYPE) && (pInfo->majorFmtVer >= 2))) {
                TULong fakeHeaderSize = SIZE_FIXED_HDR_L;
                errCode = LdrfFileWriteLong(pInfo, &fakeHeaderSize);
            } else {
                TUShort fakeHeaderSize = SIZE_FIXED_HDR_S;
                errCode = LdrfFileWriteShort(pInfo, &fakeHeaderSize);
            }
        }

        if (!errCode) {
            TUShort fakeHeaderCRC = 0x1234;
            errCode = LdrfFileWriteShort(pInfo, &fakeHeaderCRC);
        }

        if (!errCode) {
            errCode = LdrfFileWriteLong(pInfo, &(pInfo->dataAddress));
        }

        if (!errCode) {
            errCode = LdrfFileWriteLong(pInfo, &(pInfo->dataSize));
        }
    }

    return errCode;
} /* end of: LdrfWriteFileHeader */

LDRF_FN LdrfOpenFile(LPCSTR path, TLdrfFileType fileType, TUShort majorVersion, TBool checkCRC, PLdrfFileInfo* ppInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfFileInfo pInfo = NULL;
    // FILE * pFile = NULL;
    PCacheControl pCache = NULL;

    if (!ppInfo) {
        errCode = LDRF_ERR_PARAM;
    } else {
        if (path == NULL || strlen(path) > MAX_PATH) {
            // EPR 55195: If the path value is null or greater than maximum path
            // length on Windows, return file not found error.
            // For VB apps, an uninitialized string translates into a null value;
            // so this check will catch those errors (added for Distech Controls).
            // But for C/C++ apps, passing an uninitialized string will most likely 
            // cause a memory violation and crash the application.
            errCode = LDRF_ERR_NOT_FOUND;
        } else {
            pCache = CacheOpen(path, F_ACCESS_READ_BINARY, 0, 1, 1, 1, 1, F);

            // if (!pFile) {
            if ((!pCache) || (pCache->cacheErrStatus)) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                errCode = LdrfNewFileInfo(&pInfo, fileType);

                if (!errCode) {
                    // pInfo->pFile = pFile;
                    pInfo->pCache = pCache;

                    switch (fileType) {
                        case LDRF_LANG_RESOURCE:
                            errCode = LdrfReadLangFileHeader(pInfo, checkCRC);
                            break;

                        case LDRF_TYPE:
                            errCode = LdrfReadTypeFileHeader(pInfo, checkCRC);
                            break;

                        case LDRF_FPT:
                            errCode = LdrfReadFPTFileHeader(pInfo, checkCRC);
                            break;

                        default:
                            errCode = LDRF_ERR_FILE_TYPE;
                    }
                }

                if ((!errCode) && (majorVersion > pInfo->majorDataVer)) {
                    errCode = LDRF_ERR_VERSION;
                }
            }

            if (errCode) {
                // if (pFile) {
                if (pCache) {
                    // fclose(pFile);
                    CacheClose(pCache, '\0', 1);
                }

                if (pInfo) {
                    // pInfo->pFile = NULL;
                    pInfo->pCache = NULL;
                    LdrfFreeFileInfo(&pInfo);
                }
            } else {
                * ppInfo = pInfo;
            }
        }
    }

    return errCode;
} /* end of: LdrfOpenFile */

LDRF_FN LdrfEnableEmptyEntries(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if (!errCode) {
        pInfo->emptyRecordAware = T;
    }

    return errCode;
} /* end of: LdrfEnableEmptyEntries */

LDRF_FN LdrfEnableExtendedSNVTID(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if (!errCode) {
        pInfo->extendedSNVTIDaware = T;
    }

    return errCode;
} /* end of: LdrfEnableExtendedSNVTID */

LDRF_FN LdrfGetFileHdrInfo(PLdrfFileInfo pInfo, PBool pUser, LPSTR pDesc, PUShort pDescLen, LPSTR pCreator, PUShort pCreLen, LPSTR pURL, PUShort pURLLen,
    PUByte pResDescScope, PULong pResDescIndex, PUByte pResCreScope, PULong pResCreIndex)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if (!errCode) {
        if (pUser) {
            * pUser = pInfo->userFlag;
        }

        if (pDesc) {
            errCode = LdrfTransferStringOut(pDesc, pDescLen, pInfo->pDesc);

            if ((!errCode) && (pDesc)) {
                char* p = strchr(pDesc, '\032');

                if (p) {
                    * p = '\0';
                }
            }
        }
    }

    if ((!errCode) && (pCreator)) {
        errCode = LdrfTransferStringOut(pCreator, pCreLen, pInfo->pCreator);
    }

    if ((!errCode) && (pURL)) {
        errCode = LdrfTransferStringOut(pURL, pURLLen, pInfo->pURL);
    }

    if (pResDescScope) {
        * pResDescScope = pInfo->resDescScope;
    }

    if (pResDescIndex) {
        * pResDescIndex = pInfo->resDescIndex;
    }

    if (pResCreScope) {
        * pResCreScope = pInfo->resCreScope;
    }

    if (pResCreIndex) {
        * pResCreIndex = pInfo->resCreIndex;
    }

    return errCode;
} /* end of: LdrfGetFileHdrInfo */

LDRF_FN LdrfSetFileHdrInfo(PLdrfFileInfo pInfo, LPCSTR creator, LPCSTR phone, LPCSTR webid, LPCSTR URL, TUByte resDescScope, TULong resDescIndex,
    TUByte resCreScope, TULong resCreIndex)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if ((!errCode) && (!(pInfo->editAllowed))) {
        errCode = LDRF_ERR_NO_ACCESS;
    }

    if (!errCode) {
        TUShort spaceLeft = MAX_FILE_INFO_STRING - 1;

        if ((phone) && (*phone == EOS)) {
            phone = NULL;
        }

        if ((webid) && (*webid == EOS)) {
            webid = NULL;
        }
        strcpy(pInfo->pCreator, "This file was created & is maintained by ");
        spaceLeft -= (strlen(pInfo->pCreator) - 1);
        strncat(pInfo->pCreator, creator, spaceLeft);

        if (strlen(creator) < spaceLeft) {
            spaceLeft -= strlen(creator);

            if ((phone) || (webid)) {
#define CONTACT_US_AT ".\nContact us at "

                strncat(pInfo->pCreator, CONTACT_US_AT, spaceLeft);

                if (strlen(CONTACT_US_AT) < spaceLeft) {
                    spaceLeft -= strlen(CONTACT_US_AT);
                } else {
                    spaceLeft = 0;
                }

                if ((spaceLeft) && (phone)) {
                    strncat(pInfo->pCreator, phone, spaceLeft);

                    if (strlen(phone) < spaceLeft) {
                        spaceLeft -= strlen(phone);

                        if (webid) {
#define OR_AT ",\nor at "

                            strncat(pInfo->pCreator, OR_AT, spaceLeft);

                            if (strlen(OR_AT) < spaceLeft) {
                                spaceLeft -= strlen(OR_AT);
                            } else {
                                spaceLeft = 0;
                            }
                        }
                    } else {
                        spaceLeft = 0;
                    }
                }

                if ((spaceLeft) && (webid)) {
                    strncat(pInfo->pCreator, webid, spaceLeft);

                    if (strlen(webid) < spaceLeft) {
                        spaceLeft -= strlen(webid);
#define ON_THE_INTERNET " on the internet"

                        strncat(pInfo->pCreator, ON_THE_INTERNET, spaceLeft);

                        if (strlen(ON_THE_INTERNET) < spaceLeft) {
                            spaceLeft -= strlen(ON_THE_INTERNET);
                        } else {
                            spaceLeft = 0;
                        }
                    } else {
                        spaceLeft = 0;
                    }
                }
            }

            if (spaceLeft >= 2) {
                strncat(pInfo->pCreator, ".\n", spaceLeft);
            }
        }

        if (URL) {
            strncpy(pInfo->pURL, URL, MAX_FILE_INFO_STRING);
            pInfo->pURL[MAX_FILE_INFO_STRING - 1] = EOS;
        } else {
            pInfo->pURL[0] = EOS;
        }
        pInfo->resDescScope = resDescScope;
        pInfo->resDescIndex = resDescIndex;
        pInfo->resCreScope = resCreScope;
        pInfo->resCreIndex = resCreIndex;
        pInfo->needsHdrInfo = F;
        pInfo->updateHeader = T;
    }

    return errCode;
} /* end of: LdrfSetFileHdrInfo */

LDRF_FN LdrfGetFileVersion(PLdrfFileInfo pInfo, PUByte pMajorFmtVer, PUByte pMinorFmtVer, PUShort pMajorDataVer, PUByte pMinorDataVer, PUShort pYear,
    PUByte pMonth, PUByte pDay, PUByte pHour, PUByte pMinute, PUByte pSecond, PUByte pScope, PUByteArray* ppRefID)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if (!errCode) {
        if (pMajorFmtVer) {
            * pMajorFmtVer = pInfo->majorFmtVer;
        }

        if (pMinorFmtVer) {
            * pMinorFmtVer = pInfo->minorFmtVer;
        }

        if (pMajorDataVer) {
            * pMajorDataVer = pInfo->majorDataVer;
        }

        if (pMinorDataVer) {
            * pMinorDataVer = pInfo->minorDataVer;
        }

        if (pYear) {
            * pYear = pInfo->year;
        }

        if (pMonth) {
            * pMonth = pInfo->month;
        }

        if (pDay) {
            * pDay = pInfo->day;
        }

        if (pHour) {
            * pHour = pInfo->hour;
        }

        if (pMinute) {
            * pMinute = pInfo->minute;
        }

        if (pSecond) {
            * pSecond = pInfo->second;
        }

        if (pScope) {
            * pScope = pInfo->scope;
        }

        if (ppRefID) {
            PUByteArray pBA = NEW_ARRAY(TUByte, NUM_PROG_ID_BYTES);

            if (!pBA) {
                errCode = LDRF_ERR_SYS;
            } else {
                * ppRefID = pBA;
                memcpy(pBA, pInfo->refID, NUM_PROG_ID_BYTES);
            }
        }
    }

    return errCode;
} /* end of: LdrfGetFileVersion */

LDRF_FN LdrfSetFileVersion(PLdrfFileInfo pInfo, TUShort majorDataVer, TUByte minorDataVer, TUByte scope, PUByteArray pRefID) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);
    LPCSTR pDescFileType = NULL;

    if ((!errCode) && (!(pInfo->editAllowed))) {
        errCode = LDRF_ERR_NO_ACCESS;
    }

    if (!errCode) {
        switch (pInfo->fileType) {
            case LDRF_CATALOG:
                pDescFileType = "Catalog of Device Resource Files (CAT)";
                break;

            case LDRF_LANG_RESOURCE:
                pDescFileType = "Language Resource (RES)";
                break;

            case LDRF_TYPE:
                pDescFileType = "NV and CP Type and Enum (TYP)";
                break;

            case LDRF_FPT:
                pDescFileType = "Functional Profile Template (FPT)";
                break;

            default:
                errCode = LDRF_ERR_FILE_TYPE;
        }
    }

    if (!errCode) {
        if ((pInfo->majorDataVer > majorDataVer) || ((pInfo->majorDataVer == majorDataVer) && (pInfo->minorDataVer > minorDataVer))) {
            errCode = LDRF_ERR_VERSION;
        } else if (((scope == 0) && (pInfo->userFlag)) || ((scope != 0) && (!(pInfo->userFlag))) || (!pRefID)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            TUShort descLen = 0;
            pInfo->majorDataVer = majorDataVer;
            pInfo->minorDataVer = minorDataVer;
            strcpy(pInfo->pDesc + descLen, "\r\n");
            descLen += strlen(pInfo->pDesc + descLen);
            strcpy(pInfo->pDesc + descLen, pDescFileType);
            descLen += strlen(pInfo->pDesc + descLen);
            strcpy(pInfo->pDesc + descLen, " file, Data version ");
            descLen += strlen(pInfo->pDesc + descLen);
            _itoa(pInfo->majorDataVer, pInfo->pDesc + descLen, 10);
            descLen += strlen(pInfo->pDesc + descLen);
            pInfo->pDesc[descLen] = '.';
            ++descLen;

            if (pInfo->minorDataVer < 10) {
                pInfo->pDesc[descLen] = '0';
                ++descLen;
            }
            _itoa(pInfo->minorDataVer, pInfo->pDesc + descLen, 10);
            descLen += strlen(pInfo->pDesc + descLen);
            strcpy(pInfo->pDesc + descLen, ".\r\nThis file is a binary (non-ASCII) file - DO NOT EDIT!\r\n\032");
            pInfo->scope = scope;

            if (pRefID) {
                memcpy(pInfo->refID, pRefID, NUM_PROG_ID_BYTES);
            }

            {
                struct tm* ptm;
                time_t ltime;
                time(&ltime);
                ptm = gmtime(&ltime);
                pInfo->year = ptm->tm_year + 1900;
                pInfo->month = ptm->tm_mon + 1;
                pInfo->day = ptm->tm_mday;
                pInfo->hour = ptm->tm_hour;
                pInfo->minute = ptm->tm_min;
                pInfo->second = ptm->tm_sec;
            }
            pInfo->needsVerInfo = F;
            pInfo->updateHeader = T;
        }
    }

    return errCode;
} /* end of: LdrfSetFileVersion */

LDRF_FN LdrfCloseFile(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_WILD_CARD);

    if (!errCode) {
        // if (pInfo->pFile) {
        if (pInfo->pCache) {
            TLdrfErrCodes ec2 = LDRF_ERR_NONE;

            if (pInfo->fileType == LDRF_TYPE) {
                errCode = LdrfDeselectEnumSet(pInfo);
            }

            if ((!errCode) && (pInfo->updateHeader)) {
                switch (pInfo->fileType) {
                    case LDRF_CATALOG:
                        errCode = LdrfWriteCatalogHeader(pInfo);
                        break;
#ifdef LCADRF_WRITE

                    case LDRF_LANG_RESOURCE:
                        errCode = LdrfWriteLangFileHeader(pInfo);
                        break;

                    case LDRF_TYPE:
                        errCode = LdrfWriteTypeFileHeader(pInfo);
                        break;

                    case LDRF_FPT:
                        errCode = LdrfWriteFPTFileHeader(pInfo);
                        break;

#endif

                    default:
                        errCode = LDRF_ERR_UNIMPL;
                        break;
                }
            }

            if ((!errCode) && (pInfo->updateData)) {
                errCode = LdrfUpdateDataCRC(pInfo);
            }

            // if (fclose(pInfo->pFile)) {
            if (!CacheClose(pInfo->pCache, '\0', 1)) {
                ec2 = LDRF_ERR_SYS;
            }

            if (!errCode) {
                errCode = ec2;
            }
        }
        // pInfo->pFile = NULL;
        pInfo->pCache = NULL;

        LdrfFreeFileInfo(&pInfo);
    }

    return errCode;
} /* end of: LdrfCloseFile */

LDRF_FN LdrfSupportedFormats(TLdrfFileType fileType, PUByte pMajFmtLow, PUByte pMajFmtHigh) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TUByte majFmtLow = 0, majFmtHigh = 0;

    switch (fileType) {
        case LDRF_CATALOG:
            majFmtLow = 1;
            majFmtHigh = 2;
            break;

        case LDRF_TYPE:
            majFmtLow = 2;
            majFmtHigh = 6;
            break;

        case LDRF_LANG_RESOURCE:
            majFmtLow = 1;
            majFmtHigh = 3;
            break;

        case LDRF_FPT:
            majFmtLow = 1;
            majFmtHigh = 5;
            break;

        default:
            errCode = LDRF_ERR_PARAM;
    }

    if (pMajFmtLow) {
        * pMajFmtLow = majFmtLow;
    }

    if (pMajFmtHigh) {
        * pMajFmtHigh = majFmtHigh;
    }

    return errCode;
} /* end of: LdrfSupportedFormats */

LDRF_FN LdrfGetDRFAPIErrorString(TLdrfErrCodes errCodeIn, LPSTR pErrorString, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

#define FIRST_ERROR 1

    static LPCSTR errors [] =
        {
            "Unknown error code [DRF#", "Parameter error [DRF#1]", "Incorrect file type [DRF#2]", "CRC fails [DRF#3]", "Not found [DRF#4]",
        "File info handle not valid [DRF#5]", "System error [DRF#6]", "String truncated [DRF#7]", "Stale catalog [DRF#8]", "Data version error [DRF#9]",
        "Format version not supported [DRF#10]", "New file [DRF#11]", "Write error [DRF#12]", "Modification not permitted [DRF#13]", "File is full [DRF#14]",
        "Duplicate not permitted [DRF#15]", "File is not a catalog [DRF#16]", "File is not a language resource file [DRF#17]",
        "File is not a TYPE file [DRF#18]", "File is not a FPT file [DRF#19]", "File is not a format (FMT) file [DRF#20]",
        "Type tree handle not valid [DRF#21]", "Incomplete file operation or incomplete tree operation [DRF#22]",
        "Tree construction or tree scan call sequence error [DRF#23]", "Nothing is/was selected [DRF#24]", "No more nodes to scan [DRF#25]",
        "Internal error [DRF#26]", "Unimplemented feature [DRF#27]", "String service error [DRF#28]", "Empty record condition [DRF#29]",
        };
#define LAST_ERROR ((sizeof(errors)/sizeof(LPCSTR))-1)

    if ((!pErrorString) || (!pLength) || (pLength == 0)) {
        errCode = LDRF_ERR_PARAM;
    } else if ((errCodeIn >= FIRST_ERROR) && (errCodeIn <= LAST_ERROR)) {
        LPCSTR pErr = errors[errCodeIn];
        TUShort errLen = strlen(pErr);

        if (*pLength > errLen) {
            strcpy(pErrorString, pErr);
            * pLength = errLen + 1; /* Include zero termination byte */
        } else {
            strncpy(pErrorString, pErr, (* pLength)-1);
            pErrorString[(* pLength)-1] = '\0';
            errCode = LDRF_ERR_TRUNC;
        }
    } else {
        LPCSTR pErr = errors[0];
        TUShort errLen = strlen(pErr);
        char buf[10];
        _itoa(errCodeIn, buf, 10);
        strcat(buf, "]");
        errLen += strlen(buf);

        if (*pLength > errLen) {
            strcpy(pErrorString, pErr);
            strcat(pErrorString, buf);
            * pLength = errLen + 1; /* Include zero termination byte */
        } else if (*pLength > (strlen(buf) + 5)) {
            strcpy(pErrorString, "[DRF#");
            strcpy(pErrorString + 5, buf);
            errCode = LDRF_ERR_TRUNC;
        } else {
            strncpy(pErrorString, pErr, (* pLength)-1);
            pErrorString[(* pLength)-1] = '\0';
            errCode = LDRF_ERR_TRUNC;
        }
    }

    return errCode;
} /* end of: LdrfGetDRFAPIErrorString */

LDRF_FN LdrfGetDRFAPIVersion(PUShort pMajor, PUShort pMinor, PUShort pFix) {
    if (pMajor) {
        * pMajor = RELEASE_NUMBER_MAJOR;
    }

    if (pMinor) {
        * pMinor = 10*RELEASE_NUMBER_MINOR1+RELEASE_NUMBER_MINOR2;
    }

    if (pFix) {
        * pFix = 10*RELEASE_NUMBER_BUILD1+RELEASE_NUMBER_BUILD2;
    }

    return LDRF_ERR_NONE;
} /* end of: LdrfGetDRFAPIVersion */

LDRF_FN LdrfGetNextSupportedNVTType(PLdrfFileInfo pInfo, PNVTType pNVTType) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TNVTType nvtType = NVT_TYPE_UNKNOWN;

    if (!pNVTType) {
        errCode = LDRF_ERR_PARAM;
    } else {
        nvtType = *pNVTType;

        if ((nvtType < NVT_TYPE_UNKNOWN) || (nvtType >= FIRST_UNDEFINED_TYPE)) {
            errCode = LDRF_ERR_PARAM;
        }
    }

    if (pInfo) {
        errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    }

    if (!errCode) {
        ++nvtType;

        if ((nvtType > LAST_UNCONDITIONAL_TYPE) && (pInfo)) {
            if ((nvtType == NVT_TYPE_UNSIGNED_QUAD) && (!(pInfo->u.type.dataTypeUnsignedQuadAware))) {
                ++nvtType;
            }

            if ((nvtType == NVT_TYPE_DOUBLE_FLOAT) && (!(pInfo->u.type.dataTypeDoubleFloatAware))) {
                ++nvtType;
            }
            if ((nvtType == NVT_TYPE_SIGNED_INT64) && (!(pInfo->u.type.dataTypeSignedInt64Aware))) {
                ++nvtType;
            }
            if ((nvtType == NVT_TYPE_UNSIGNED_INT64) && (!(pInfo->u.type.dataTypeUnsignedInt64Aware))) {
                ++nvtType;
            }
        }

        if (nvtType >= FIRST_UNDEFINED_TYPE) {
            nvtType = NVT_TYPE_UNKNOWN;
        }
    }

    if (!errCode) {
        * pNVTType = nvtType;
    }

    return errCode;
} /* end of: LdrfGetNextSupportedNVTType */

#define FIRST_TYPE_NAME 1

static LPCSTR typeNames [] =
    {
        "Unknown type", "signed char", "unsigned char", "signed short int", "unsigned short int", "signed long int", "unsigned long int", "enum", "array",
    "struct", "union", "bitfield", "float", "signed quad int", "reference", "unsigned quad int", "double float", "signed int64", "unsigned int64",
    };
#define LAST_TYPE_NAME ((sizeof(typeNames)/sizeof(LPCSTR))-1)

LDRF_FN LdrfGetTypeNameString(TNVTType nvtType, LPSTR pTypeNameString, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    if ((!pTypeNameString) || (!pLength) || (pLength == 0) || (nvtType < NVT_TYPE_UNKNOWN) || (nvtType >= FIRST_UNDEFINED_TYPE)) {
        errCode = LDRF_ERR_PARAM;
    } else {
        LPCSTR pTypeName = typeNames[nvtType];
        TUShort nameLen = strlen(pTypeName);

        if (*pLength > nameLen) {
            strcpy(pTypeNameString, pTypeName);
            * pLength = nameLen + 1; /* Include zero termination byte */
        } else {
            strncpy(pTypeNameString, pTypeName, (* pLength)-1);
            pTypeNameString[(* pLength)-1] = '\0';
            errCode = LDRF_ERR_TRUNC;
        }
    }

    return errCode;
} /* end of: LdrfGetTypeNameString */

LDRF_FN LdrfLookupTypeNameString(LPCSTR pTypeNameString, PNVTType pNVTType) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NOT_FOUND;
    TNVTType nvtType;

    if ((!pTypeNameString) || (!pNVTType)) {
        errCode = LDRF_ERR_PARAM;
    } else {
        for ( nvtType = FIRST_TYPE_NAME; nvtType <= LAST_TYPE_NAME; ++nvtType ) {
            if (!STRICMP(pTypeNameString, typeNames[nvtType])) {
                errCode = LDRF_ERR_NONE;
                * pNVTType = nvtType;
                break;
            }
        }
    }

    return errCode;
} /* end of: LdrfLookupTypeNameString */
