/*
 * DRFLANG.C - This is the Resource File API file responsible for 
 * language resource file access methods
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

TLdrfErrCodes LdrfReadLangFileHeader(PLdrfFileInfo pInfo, TBool checkCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfReadFileHeader(pInfo, checkCRC);

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, pInfo->refID, sizeof(pInfo->refID));
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &(pInfo->scope), sizeof(pInfo->scope));
    }

    if (!errCode) {
        errCode = LdrfFileReadLong(pInfo, &(pInfo->u.lang.locale));
    }

    if (!errCode) {
        errCode = LdrfFileRead24(pInfo, &(pInfo->u.lang.numStrings));
        pInfo->u.lang.capacityStrings = pInfo->u.lang.numStrings;
    }

    if (pInfo->majorFmtVer > 2) {
        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &(pInfo->u.lang.numEmptyStrings));
        }
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.lang.numStrings, pInfo->u.lang.numStrings, &(pInfo->u.lang.pDirStrings));
    }

    if (!errCode) {
        pInfo->u.lang.needsLangFileInfo = F;
        pInfo->u.lang.pHashTbl = NULL;
    }

    return errCode;
} /* end of: LdrfReadLangFileHeader */

TLdrfErrCodes LdrfFreeLangFileInfo(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    FREE(pInfo->u.lang.pDirStrings);

    if (pInfo->u.lang.pHashTbl) {
#ifndef LCADRF_WRITE

        errCode = LDRF_ERR_INTERNAL;
#else

        TULong index;

        for ( index = 0; index <= 65535; ++index ) {
            PHashTblEntry pTemp, pEntry = pInfo->u.lang.pHashTbl->pEntry[index];

            while (pEntry) {
                pTemp = pEntry;
                pEntry = pEntry->pNextEntry;
                FREE(pTemp);
            }
        }
        FREE(pInfo->u.lang.pHashTbl);

#endif

    }

    return errCode;
} /* end of: LdrfFreeLangFileInfo */

LDRF_FN LdrfGetLangFileInfo(PLdrfFileInfo pInfo, PULong pLocale, PULong pNumResources) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_LANG_RESOURCE);

    if (!errCode) {
        if (pLocale) {
            * pLocale = pInfo->u.lang.locale;
        }

        if (pNumResources) {
            * pNumResources = pInfo->u.lang.numStrings;
        }
    }

    return errCode;
} /* end of: LdrfGetLangFileInfo */

LDRF_FN LdrfGetResourceString(PLdrfFileInfo pInfo, TULong index, LPSTR pString, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateResourceString(pInfo, index);

    if ((!errCode) || (errCode == LDRF_ERR_EMPTY_RECORD)) {
        if ((!pString) || (!pLength)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            TUShort len;

            if (errCode == LDRF_ERR_EMPTY_RECORD) {
                if (!(pInfo->emptyRecordAware)) {
                    char nameBuf[20];
                    strcpy(nameBuf, "empty");
                    _itoa(index, nameBuf + 5, 10);
                    strcat(nameBuf, "~");
                    len = strlen(nameBuf);

                    if (len > *pLength) {
                        strncpy(pString, nameBuf, (* pLength)-1);
                        pString[(* pLength)-1] = '\0';
                        errCode = LDRF_ERR_TRUNC;
                    } else {
                        strcpy(pString, nameBuf);
                        errCode = LDRF_ERR_NONE;
                    }
                    * pLength = len;
                }
            } else {
                errCode = LdrfReadLangRecStart(pInfo, pInfo->u.lang.pDirStrings[index], NULL, &len, NULL, NULL);

                if (!errCode) {
                    TLdrfErrCodes ec2;

                    if (*pLength < len) {
                        len = *pLength - 1;
                        errCode = LDRF_ERR_TRUNC;
                    } else {
                        * pLength = len; /* Includes termination zero byte */
                    }
                    ec2 = LdrfFileReadBytes(pInfo, pString, len);

                    if (ec2) {
                        errCode = ec2;
                    } else if (errCode == LDRF_ERR_TRUNC) {
                        pString[len] = EOS;
                    }
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfGetResourceString */

LDRF_FN LdrfValidateResourceString(PLdrfFileInfo pInfo, TULong index) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_LANG_RESOURCE);

    if (!errCode) {
        if (index < 1) {
            errCode = LDRF_ERR_PARAM;
        } else if (index > pInfo->u.lang.numStrings) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (!(pInfo->u.lang.pDirStrings)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.lang.pDirStrings[index])) {
            errCode = LDRF_ERR_EMPTY_RECORD;
        }
    }

    return errCode;
} /* end of: LdrfValidateResourceString */

LDRF_FN LdrfFindEmptyResourceString(PLdrfFileInfo pInfo, PULong pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_LANG_RESOURCE);
    TULong index = 0;

    if (!errCode) {
        if (!(pInfo->u.lang.numStrings)) {
            index = 1;
        } else if (!(pInfo->u.lang.pDirStrings)) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode = LDRF_ERR_NOT_FOUND;

            for ( index = 1; index <= pInfo->u.lang.numStrings; ++index ) {
                if (!(pInfo->u.lang.pDirStrings[index])) {
                    errCode = LDRF_ERR_NONE;
                    break;
                }
            }

            if (errCode) {
                if (pInfo->u.lang.numStrings < ((65536UL*256)-1)) {
                    index = pInfo->u.lang.numStrings + 1;
                    errCode = LDRF_ERR_NONE;
                }
            }
        }
    }

    if (pIndex) {
        * pIndex = index;
    }

    return errCode;
} /* end of: LdrfFindEmptyResourceString */

TLdrfErrCodes LdrfReadLangRecStart(PLdrfFileInfo pInfo, TULong startAddress, PULong pIndex, PUShort pStrLen, PUShort pHashCode, PUByte pUnicode) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;
    TUShort recSize;

    if (startAddress) {
        if (pIndex) {
            TULong index;
            errCode = LdrfFileSeek(pInfo, startAddress);

            if (!errCode) {
                errCode = LdrfFileRead24(pInfo, &index);

                if (!errCode) {
                    * pIndex = index;
                }
            }
        } else {
            errCode = LdrfFileSeek(pInfo, startAddress + OFFSET_LANG_FILE_REC_SIZE);
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &recSize);
    }

    if (!errCode) {
        TUShort hashCode;
        errCode = LdrfFileReadShort(pInfo, &hashCode);

        if (!errCode) {
            recSize -= sizeof(hashCode);

            if (pHashCode) {
                * pHashCode = hashCode;
            }
        }
    }

    if (!errCode) {
        TUByte unicode;
        errCode = LdrfFileReadBytes(pInfo, &unicode, sizeof(unicode));

        if (!errCode) {
            recSize -= sizeof(unicode);

            if (unicode) {
                errCode = LDRF_ERR_UNIMPL;
            } else if (pUnicode) {
                * pUnicode = unicode;
            }
        }
    }

    if ((!errCode) && pStrLen) {
        * pStrLen = recSize;
    }

    return errCode;
} /* end of: LdrfReadLangRecStart */

typedef struct sLangInfoElement {
    /* The "key" is the index PLUS ONE! */
    TUByte extension[4];
    TULong langStringIndex;
    TULong locale;
    TULong MSLocaleID;
} TLangInfoElement;

typedef struct sLangInfo {
    TUShort numLanguages;
    TLangInfoElement langInfo[1];
} TLangInfo;

typedef TLangInfo* PLangInfo;

TLdrfErrCodes LdrfGetInfFilename(LPCSTR infDirectory, TCHAR* buf, DWORD* pBufLen) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    char* p;

    if ((!infDirectory) || (*infDirectory == '\0')) {
        errCode = LdrfReadRegistryKeyString(HKEY_CLASSES_ROOT, TEXT("CLSID\\{1CB4B4E3-EC74-11CF-8CC7-444553540000}\\InprocServer32"), NULL, buf, pBufLen);

        if (!errCode) {
            p = strstr(_strlwr(buf), ".dll");

            if (!p) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                strcpy(p, ".inf");
            }
        }
    } else {
        strncpy(buf, infDirectory, *pBufLen);
        strncat(buf, DIR_SEPARATOR_STRING "lcadrf32.inf", * pBufLen);
        p = strstr(buf, ".inf");

        if (!p) {
            errCode = LDRF_ERR_TRUNC;
        }
    }

    return errCode;
} /* end of: LdrfGetInfFilename */


// Checks whether a LonWorks Registry entry exists for the specified path name
// (typically either "LonWorks Path" or "LonWorks Data Path").  If the entry
// exists, a path to the "Types" sub-directory is returned.  Optionally, tests
// whether a catalog ("LDRF.CAT") exists in the Types sub-directory.
static TLdrfErrCodes
CheckCatalogPath(LPCSTR lonWorksPathName, LPTSTR typesPath, int typesPathSize, TBool checkExists)
{
    // fetch the specified LonWorks path from the Registry
    int size = typesPathSize;
    TLdrfErrCodes errCode = LdrfReadRegistryKeyString(HKEY_LOCAL_MACHINE, TEXT("Software\\LonWorks"), lonWorksPathName, typesPath, &size);
    if (errCode)
        // Registry setting doesn't exist
        return errCode;

    if (!size)
        // LonWorks path is empty
        return LDRF_ERR_EMPTY_RECORD;

    // build the path to the Types directory
    strcat(typesPath, DIR_SEPARATOR_STRING "Types");
    if (!strstr(typesPath, DIR_SEPARATOR_STRING "Types"))
        return LDRF_ERR_TRUNC;

    // check whether the catalog actually exists
    // (some installers create the Registry entry
    //  but don't move the catalog, e.g. OpenLDV)
    if (checkExists)
    {
        struct _stat stat_buffer = {0};
        TCHAR catalogPath[MAX_PATH + 1];

        // build the path to the catalog
        strcpy(catalogPath, typesPath);
        strcat(catalogPath, DIR_SEPARATOR_STRING CATALOG_NAME);
        if (!strstr(catalogPath, DIR_SEPARATOR_STRING CATALOG_NAME))
            return LDRF_ERR_TRUNC;

        if (_stat(catalogPath, &stat_buffer) != ERROR_SUCCESS)
            // file does not exist
            return LDRF_ERR_NOT_FOUND;
    }

    return LDRF_ERR_NONE;
}

// This function redirects catalog paths if they are set to the old standard 
// "<lonworks path>\Types", to the new standard "<lonworks data path>\Types" ,
// if the setting and file exists.
// The result, whether it is redirected or not, is returned in redirectedPath.
TLdrfErrCodes LdrfRedirectCatalogPath(LPCSTR specifiedPath, LPSTR redirectedPath, DWORD pathSize)
{
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    DWORD size;
	TBool  redirectFlag = F;

    TCHAR LongFormSpecifiedPath[MAX_PATH + 1];
	TCHAR LongFormPathToBeRedirected[MAX_PATH + 1];
    TCHAR PathToBeRedirected[MAX_PATH + 1];

	if (specifiedPath && *specifiedPath != '\0')
    {
		// See if the specified directory is "<lonworks path>\Types" directory. If it is, redirect to the
		// "<lonworks data path>\Types" directory.
		size = MAX_PATH;
		PathToBeRedirected[0] = '\0';
		if (!LdrfReadRegistryKeyString(HKEY_LOCAL_MACHINE, TEXT("Software\\LonWorks"), TEXT("LonWorks Path"), PathToBeRedirected, &size))
			strncat(PathToBeRedirected, DIR_SEPARATOR_STRING "Types", pathSize);

		// Based on application runtime component behavior, a particular app might pass us an
		// 8.3 format path string, OR one might be saved to the Registry.  To make this check,
		// we will make an OS call to get long (not 8.3) pathnames for the path passed in, and 
		// the one that we get from the Registry.
		LongFormPathToBeRedirected[0] = '\0';
		size = GetLongPathName(PathToBeRedirected, LongFormPathToBeRedirected, MAX_PATH);
		if (size == 0)
			// these failure cases should not happen, but if so, use potentially short path instead
			strncpy(LongFormPathToBeRedirected, PathToBeRedirected, MAX_PATH);
		else if (size > MAX_PATH)
			// we have an error code for super-long path length problems
			errCode = LDRF_ERR_TRUNC;

		LongFormSpecifiedPath[0] = '\0';
		size = GetLongPathName(specifiedPath, LongFormSpecifiedPath, MAX_PATH);
		if (size == 0)
			// these failure cases should not happen, but if so, use potentially short path instead
			strncpy(LongFormSpecifiedPath, specifiedPath, MAX_PATH);
		else if (size > MAX_PATH)
		    // we have an error code for super-long path length problems
			errCode = LDRF_ERR_TRUNC;

		// if the specified path ends in an optional DIR_SEPARATOR_CHAR, delete it before compare
		if (strlen(LongFormSpecifiedPath) > 0 &&
            LongFormSpecifiedPath[strlen(LongFormSpecifiedPath) - 1] == DIR_SEPARATOR_CHAR)
				LongFormSpecifiedPath[strlen(LongFormSpecifiedPath) - 1] = '\0';

		if (_strcmpi(LongFormSpecifiedPath, LongFormPathToBeRedirected) == 0)
			// Trying to access the catalog in LonWorks installation folder.  Should get this from
            // the LonWorks data tree instead
            redirectFlag = T;
	}

	if (errCode)
        return errCode;

	if (!specifiedPath || *specifiedPath == '\0' || redirectFlag) 
	{
        // use existing LWDP catalog, if any
        errCode = CheckCatalogPath(TEXT("LonWorks Data Path"), redirectedPath, pathSize, T);
        if (errCode)
            // use existing LWP catalog, if any
            errCode = CheckCatalogPath(TEXT("LonWorks Path"), redirectedPath, pathSize, T);
        if (errCode)
            // use (new) LWDP catalog, if possible
            errCode = CheckCatalogPath(TEXT("LonWorks Data Path"), redirectedPath, pathSize, F);
        if (errCode)
            // use (new) LWP catalog, if possible
            errCode = CheckCatalogPath(TEXT("LonWorks Path"), redirectedPath, pathSize, F);
        return errCode;
	} 

	strncpy(redirectedPath, specifiedPath, pathSize);
	if (strlen(specifiedPath) > pathSize) 
		return LDRF_ERR_TRUNC;

    return LDRF_ERR_NONE;
}

TLdrfErrCodes LdrfGetCatalog(LPCSTR catDirectory, PLdrfFileInfo* pCatInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    TCHAR buf[MAX_PATH + 1];
    PLdrfFileInfo pInfo = NULL;

    errCode = LdrfRedirectCatalogPath(catDirectory, buf, MAX_PATH);

    if (!errCode) {
        errCode = LdrfOpenCatalog(buf, T, &pInfo);

        if (errCode) {
            errCode = LdrfOpenCatalog(buf, F, &pInfo);

            if (errCode == LDRF_ERR_STALE) {
                errCode = LdrfCatalogRefresh(pInfo);
            }

            if (!errCode) {
                errCode = LdrfCloseCatalog(pInfo);
            }

            if (!errCode) {
                errCode = LdrfOpenCatalog(buf, F, &pInfo);
            }
        }
    }

    if (pCatInfo) {
        * pCatInfo = pInfo;
    }

    return errCode;
} /* end of: LdrfGetCatalog */

TLdrfErrCodes LdrfReadLangInfoFile(LPCSTR infDirectory, PLangInfo* ppLangInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLangInfo pLangInfo = NULL;
#define BUF_SIZE 256

    char buf[BUF_SIZE + 1];
    TULong bufSize = BUF_SIZE;

    errCode = LdrfGetInfFilename(infDirectory, buf, &bufSize);

    if (!errCode) {
        FILE* pFile = fopen(buf, F_ACCESS_READ_TEXT);

        if (!pFile) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else {
            int numLanguages = 0, key = 1;

            if (fgets(buf, BUF_SIZE, pFile) != NULL) {
                sscanf(buf, "%d", &numLanguages);
            }

            if (numLanguages) {
                pLangInfo = LdrfAllocFn(sizeof(TLangInfo) + (sizeof(TLangInfoElement) * (numLanguages - 1)));

                if (!pLangInfo) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    pLangInfo->numLanguages = numLanguages;
                }
            }

            while ((key <= numLanguages) && (!errCode)) {
                if (fgets(buf, BUF_SIZE, pFile) != NULL) {
                    unsigned int localeDummy, MSLocaleDummy;
                    unsigned int langStringIndexDummy;

                    if (sscanf(buf, "%c%c%c %d %x %x", &(pLangInfo->langInfo[key - 1].extension[0]), &(pLangInfo->langInfo[key - 1].extension[1]),
                        &(pLangInfo->langInfo[key - 1].extension[2]), &langStringIndexDummy, &localeDummy, &MSLocaleDummy)
                        != 6)
                        {
                        errCode = LDRF_ERR_PARAM;
                    }
                    pLangInfo->langInfo[key - 1].langStringIndex
            = langStringIndexDummy;

                    pLangInfo->langInfo[key - 1].locale = localeDummy;
                    pLangInfo->langInfo[key - 1].MSLocaleID = MSLocaleDummy;
                    pLangInfo->langInfo[key - 1].extension[3] = '\0';
                } else {
                    errCode = LDRF_ERR_PARAM;
                }
                ++key;
            }
            fclose(pFile);
        }
    }

    if (ppLangInfo) {
        * ppLangInfo = pLangInfo; // This function allocates and
    // returns a lang-info structure.  Caller must free it.
    } else {
        FREE(pLangInfo);
    }

    return errCode;
} /* end of: LdrfReadLangInfoFile */

LDRF_FN LdrfGetNumLanguages(LPCSTR infDirectory, PUShort pNumLanguages) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;
    PLangInfo pLangInfo = NULL;

    errCode = LdrfReadLangInfoFile(infDirectory, &pLangInfo);

    if (!errCode) {
        if (pNumLanguages) {
            * pNumLanguages = pLangInfo->numLanguages;
        }
    }
    FREE(pLangInfo);

    return errCode;
} /* end of: LdrfGetNumLanguages */

TLdrfErrCodes LdrfGetStandardString(LPCSTR pCatDirectory, TULong index, TULong myLocale, LPSTR pString, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode;
    PLdrfFileInfo pCatInfo, pFileInfo;
    char buf[MAX_PATH+1];

    errCode = LdrfGetCatalog(pCatDirectory, &pCatInfo);

    if (!errCode) {
        TUShort bufLen = MAX_PATH;
        TUByte progID[NUM_PROG_ID_BYTES];
        memset(progID, '\0', NUM_PROG_ID_BYTES);
        errCode = LdrfSearchCatalog(pCatInfo, progID, 0, LDRF_LANG_RESOURCE, myLocale, buf, &bufLen, NULL, NULL);
        LdrfCloseCatalog(pCatInfo);

        if (!errCode) {
            errCode = LdrfOpenFile(buf, LDRF_LANG_RESOURCE, 0, T, &pFileInfo);
        }

        if (!errCode) {
            errCode = LdrfGetResourceString(pFileInfo, index, pString, pLength);
            LdrfCloseFile(pFileInfo);
        }
    }

    return errCode;
} /* end of: LdrfGetStandardString */

LDRF_FN LdrfGetLanguageInfo(LPCSTR pInfDirectory, TULong myLocale, TUShort key, LPCSTR pCatDirectory, PULong pMSLocaleID, PULong pLocale,
    PUByteArray pFileExtension, LPSTR pString, PUShort pLength)
    {
    /* Local variables */
    TLdrfErrCodes errCode;
    PLangInfo pLangInfo;

    errCode = LdrfReadLangInfoFile(pInfDirectory, &pLangInfo);

    if (!errCode) {
        if ((key == 0) || (key > pLangInfo->numLanguages)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            if (pMSLocaleID) {
                * pMSLocaleID = pLangInfo->langInfo[key - 1].MSLocaleID;
            }

            if (pLocale) {
                * pLocale = pLangInfo->langInfo[key - 1].locale;
            }

            if (pFileExtension) {
                memcpy(pFileExtension, &(pLangInfo->langInfo[key - 1].extension), 4);
            }

            if (pString) {
                errCode = LdrfGetStandardString(pCatDirectory, pLangInfo->langInfo[key - 1].langStringIndex, myLocale, pString, pLength);
            }
        }
    }
    FREE(pLangInfo);

    return errCode;
} /* end of: LdrfGetLanguageInfo */

LDRF_FN LdrfGetLanguageKeyFromLocale(LPCSTR pInfDirectory, TULong locale, PUShort pKey) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;
    PLangInfo pLangInfo = NULL;

    errCode = LdrfReadLangInfoFile(pInfDirectory, &pLangInfo);

    if (!errCode) {
        TUShort key = 1;
        errCode = LDRF_ERR_NOT_FOUND;

        while (key <= pLangInfo->numLanguages) {
            if (pLangInfo->langInfo[key - 1].locale == locale) {
                if (pKey) {
                    * pKey = key;
                }
                errCode = LDRF_ERR_NONE;
                break;
            }
            ++key;
        }
    }
    FREE(pLangInfo);

    return errCode;
} /* end of: LdrfGetLanguageKeyFromLocale */

LDRF_FN LdrfGetLanguageKeyFromMSLocaleID(LPCSTR pInfDirectory, TULong MSLocaleID, PUShort pKey) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;
    PLangInfo pLangInfo = NULL;

    errCode = LdrfReadLangInfoFile(pInfDirectory, &pLangInfo);

    if (!errCode) {
        TUShort key = 1;
        errCode = LDRF_ERR_NOT_FOUND;

        while (key <= pLangInfo->numLanguages) {
            if (pLangInfo->langInfo[key - 1].MSLocaleID == MSLocaleID) {
                if (pKey) {
                    * pKey = key;
                }
                errCode = LDRF_ERR_NONE;
                break;
            }
            ++key;
        }
    }
    FREE(pLangInfo);

    return errCode;
} /* end of: LdrfGetLanguageKeyFromMSLocaleID */

LDRF_FN LdrfGetLanguageKeyFromExtension(LPCSTR pInfDirectory, PUByteArray fileExtension, PUShort pKey) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;
    PLangInfo pLangInfo = NULL;

    errCode = LdrfReadLangInfoFile(pInfDirectory, &pLangInfo);

    if (!errCode) {
        TUShort key = 1;
        errCode = LDRF_ERR_NOT_FOUND;

        while (key <= pLangInfo->numLanguages) {
            if (!STRICMP(pLangInfo->langInfo[key - 1].extension, fileExtension)) {
                if (pKey) {
                    * pKey = key;
                }
                errCode = LDRF_ERR_NONE;
                break;
            }
            ++key;
        }
    }
    FREE(pLangInfo);

    return errCode;
} /* end of: LdrfGetLanguageKeyFromExtension */

typedef struct sLdrfStringLocaleListElement TLdrfStringLocaleListElement;
typedef TLdrfStringLocaleListElement* PLdrfStringLocaleListElement;

struct sLdrfStringLocaleListElement {
    TULong locale;
    PLdrfStringLocaleListElement pNextLocale;
};

#define ENU_LOCALE 0x656E7500

struct sLdrfStringService {
    TULong validator;
    PLdrfStringService pNextValid;
    PLdrfFileInfo pCatInfo;
    PLdrfStringLocaleListElement pLocaleList;
    int numLocales;
    TBool localeListContainsENU;

    /* The 0th entry is used as a temporary for swapping and debugging.
     * It is always the last file that was accessed or matched in cache. */
    struct {
        PLdrfFileInfo pFileInfo;
        TUByte progID[NUM_PROG_ID_BYTES];
        TUByte scope;
        TULong locale;
    } cachedFile[STRING_SERVICE_MAX_CACHED_FILES + 1];
    TUShort numCachedFiles;
    /* The 0th entry is used as a temporary for swapping and debugging.
     * It is always the last string that was accessed or matched in cache. */
    struct {
        char* pString;
        PLdrfFileInfo pFileInfo;
        TULong index;
    } cachedString[STRING_SERVICE_MAX_CACHED_STRINGS + 1];
    TUShort numCachedStrings;
};

static PLdrfStringService pValidStringServiceList;

void LdrfLangInitialize(void) {

    pValidStringServiceList = NULL;
} /* end of: LdrfLangInitialize */

LDRF_FN LdrfStartStringService(TULong locale, LPCSTR pCatDirectory, PLdrfStringService* ppService) {
    /* Local variables */
    TLdrfErrCodes errCode;
    PLdrfStringService pService = NULL;
    PLdrfFileInfo pCatInfo;

    errCode = LdrfGetCatalog(pCatDirectory, &pCatInfo);

    if (!errCode) {
        errCode = LDRF_ERR_SYS;
        pService = NEW(TLdrfStringService);

        if (pService) {
            memset(pService, 0, sizeof(TLdrfStringService));
            pService->pNextValid = pValidStringServiceList;
            pValidStringServiceList = pService;
            pService->validator = VALIDATION_CODE_STRING_SERVICE;
            pService->pCatInfo = pCatInfo;
            errCode = LdrfAddStringServiceLocale(pService, locale);

            if (errCode) {
                LdrfStopStringService(&pService);
            }
        } else {
            LdrfCloseCatalog(pCatInfo);
        }
    }

    * ppService = pService;

    return errCode;
} /* end of: LdrfStartStringService */

TLdrfErrCodes LdrfFindStringService(PLdrfStringService pService, PLdrfStringService* ppPrevService) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_STRING_SERVICE;
    PLdrfStringService pPrevService = NULL;
    PLdrfStringService pScan = pValidStringServiceList;

    while ((pScan) && (pScan != pService)) {
        pPrevService = pScan;
        pScan = pScan->pNextValid;
    }

    if (pScan) {
        errCode = LDRF_ERR_NONE;

        if (ppPrevService) {
            * ppPrevService = pPrevService;
        }
    }

    return errCode;
} /* end of: LdrfFindStringService */

TLdrfErrCodes LdrfValidateStringServicePointer(PLdrfStringService pService) {
    /* Local variables */
    PLdrfStringService pPrevService;
    TLdrfErrCodes errCode = LdrfFindStringService(pService, &pPrevService);

    if (!errCode) {
        if ((pService->validator != VALIDATION_CODE_STRING_SERVICE) || (!(pService->pCatInfo))) {
            errCode = LDRF_ERR_STRING_SERVICE;
        }
    }

    if (!errCode) {
        /* Optimize future accesses - bring this one to the front of the list */
        if (pPrevService) {
            pPrevService->pNextValid = pService->pNextValid;
            pService->pNextValid = pValidStringServiceList;
            pValidStringServiceList = pService;
        }
    }

    return errCode;
} /* end of: LdrfValidateStringServicePointer */

LDRF_FN LdrfAddStringServiceLocale(PLdrfStringService pService, TULong locale) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;
    PLdrfStringLocaleListElement pNew, pScan, pPrev;

    if (pService) {
        errCode = LdrfValidateStringServicePointer(pService);

        if (!errCode) {
            /* Check that locale not already in list, find list end */
            pPrev = NULL;
            pScan = pService->pLocaleList;

            while (pScan != NULL) {
                if (pScan->locale == locale) {
                    errCode = LDRF_ERR_DUPLICATE;
                    break;
                }
                pPrev = pScan;
                pScan = pScan->pNextLocale;
            }

            if (!errCode) {
                errCode = LDRF_ERR_SYS;
                pNew = NEW(TLdrfStringLocaleListElement);

                if (pNew) {
                    errCode = LDRF_ERR_NONE;
                    pNew->locale = locale;

                    if (locale == ENU_LOCALE) {
                        pService->localeListContainsENU = T;
                    }
                    pNew->pNextLocale = NULL;

                    if (pPrev) {
                        pPrev->pNextLocale = pNew;
                    } else {
                        pService->pLocaleList = pNew;
                    }
                    ++(pService->numLocales);
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfAddStringServiceLocale */

LDRF_FN LdrfStopStringService(PLdrfStringService* ppService) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;
    PLdrfStringService pService, pPrevService;
    PLdrfStringLocaleListElement pLocale;
    TUShort index;

    if (ppService) {
        errCode = LDRF_ERR_NONE;
        pService = *ppService;
        * ppService = NULL;
        errCode = LdrfFindStringService(pService, &pPrevService);

        if (!errCode) {
            /* Remove this string service structure from the list */
            if (pPrevService) {
                pPrevService->pNextValid = pService->pNextValid;
            } else {
                pValidStringServiceList = pService->pNextValid;
            }

            /* Close the catalog */
            if (pService->pCatInfo) {
                errCode = LdrfCloseCatalog(pService->pCatInfo);
            }

            /* Free the locale list */
            while (pService->pLocaleList) {
                pLocale = pService->pLocaleList;
                pService->pLocaleList = pLocale->pNextLocale;
                FREE(pLocale);
            }

            /* Close cached files */
            for ( index = 1; index <= pService->numCachedFiles; ++index ) {
                TLdrfErrCodes ec2;
                ec2 = LdrfCloseFile(pService->cachedFile[index].pFileInfo);

                if (!errCode) {
                    errCode = ec2;
                }
            }

            /* Free cached strings */
            for ( index = 1; index <= pService->numCachedStrings; ++index ) {
                FREE(pService->cachedString[index].pString);
            }

            /* The service is stopped. */
            FREE(pService);
        }
    }

    return errCode;
} /* end of: LdrfStopStringService */

static TLdrfErrCodes LdrfStringServiceRequestHelper(PLdrfStringService pService, PUByteArray pProgID, TUByte scope, TULong locale, TULong stringIndex,
    TUShort stringBufLength, char* pString, PUShort pStringLength)
    {
    /* Local variables */
    char fileNameBuf[257];
    TUShort fileNameLength, stringLength, index;
    TLdrfErrCodes errCode = LDRF_ERR_NONE;


    /* Search first for file in cache */
    for ( index = 1; index <= pService->numCachedFiles; ++index ) {
        if ((!memcmp(pService->cachedFile[index].progID, pProgID, NUM_PROG_ID_BYTES)) && (pService->cachedFile[index].scope == scope)
            && (pService->cachedFile[index].locale == locale))
            {
            /* Matched entry in cache */
            break;
        }
    }

    if (index > pService->numCachedFiles) {
        // Didn't find match in cache - open a new file, use [0] as a temp
        fileNameLength = 256;
        errCode = LdrfSearchCatalog(pService->pCatInfo, pProgID, scope, LDRF_LANG_RESOURCE, locale, fileNameBuf, &fileNameLength, NULL, NULL);

        if (!errCode) {
            memcpy(pService->cachedFile[0].progID, pProgID, NUM_PROG_ID_BYTES);
            pService->cachedFile[0].scope = scope;
            pService->cachedFile[0].locale = locale;
            errCode = LdrfOpenFile(fileNameBuf, LDRF_LANG_RESOURCE, 0, T, &(pService->cachedFile[0].pFileInfo));

            if (!errCode) {
                errCode = LdrfEnableEmptyEntries(pService->cachedFile[0].pFileInfo);
            }
        }

        if (!errCode) {
            /* Put this file at end of active cache entries */
            if (pService->numCachedFiles >= STRING_SERVICE_MAX_CACHED_FILES) {
                // No inactive entry - zap a random entry!
                TUShort zap, indexStr;
                pService->numCachedFiles = STRING_SERVICE_MAX_CACHED_FILES;
                zap = (((TUShort)(rand())) % (STRING_SERVICE_MAX_CACHED_FILES - 1)) + 1;

                // When the file is zapped, any matching strings must be purged
                for ( indexStr = 1; indexStr <= pService->numCachedStrings; ++indexStr ) {
                    if (pService->cachedString[indexStr].pFileInfo == pService->cachedFile[zap].pFileInfo) {
                        FREE(pService->cachedString[indexStr].pString);

                        if (indexStr < pService->numCachedStrings) {
                            // Copy in the last active string to fill
                            // the hole of the zapped string-cache entry
                            memcpy(&(pService->cachedString[indexStr]), &(pService->cachedString[pService->numCachedStrings]),
                                sizeof(pService->cachedString[0]));
                            pService->cachedString[pService->numCachedStrings].pString = NULL;
                        } else {
                            pService->cachedString[indexStr].pString = NULL;
                        }
                        --(pService->numCachedStrings);
                    }
                }

                // Close the file of the file-cache entry being zapped
                errCode = LdrfCloseFile(pService->cachedFile[zap].pFileInfo);

                // Move the LAST file-cache entry to the zapped one, leaving
                // the LAST as an inactive entry for the new file to occupy
                index = STRING_SERVICE_MAX_CACHED_FILES;
                memcpy(&(pService->cachedFile[zap]), &(pService->cachedFile[index]), sizeof(pService->cachedFile[0]));
            } else {
                index = ++(pService->numCachedFiles);
            }

            // Copy the "new" file's info (in [0]) to the inactive entry
            // designated by [index].
            memcpy(&(pService->cachedFile[index]), &(pService->cachedFile[0]), sizeof(pService->cachedFile[0]));
        }
    } else {
        // File was already in cache - copy the found entry to [0] as temp
        memcpy(&(pService->cachedFile[0]), &(pService->cachedFile[index]), sizeof(pService->cachedFile[0]));

        if (index > 1) {
            // If the file found was past index [1] (i.e. not first),
            // then this is a "hot file" optimization that promotes files
            // getting more "hits" by swapping them with the next "higher"
            --index;
            memcpy(&(pService->cachedFile[index + 1]), &(pService->cachedFile[index]), sizeof(pService->cachedFile[0]));
            // Copy the "hit" file's info (in [0]) to the inactive entry
            // designated by [index].
            memcpy(&(pService->cachedFile[index]), &(pService->cachedFile[0]), sizeof(pService->cachedFile[0]));
        }
    }


    /* Now search for string in the string cache.
     * Note that the applicable file is always in [0]! */
    if (!errCode) {
        if ((!index) || (!(pService->cachedFile[0].pFileInfo))) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            for ( index = 1; index <= pService->numCachedStrings; ++index ) {
                if ((pService->cachedString[index].pFileInfo == pService->cachedFile[0].pFileInfo) && (pService->cachedString[index].index == stringIndex)) {
                    /* Found match in string cache */
                    break;
                }
            }

            if (index > pService->numCachedStrings) {
                /* Didn't find match in string cache - read in from file [0]
                 * and put result in string [0], used as a temp */
                stringLength = stringBufLength;
                errCode = LdrfGetResourceString(pService->cachedFile[0].pFileInfo, stringIndex, pString, &stringLength);

                if (errCode == LDRF_ERR_EMPTY_RECORD) {
                    errCode = LDRF_ERR_NOT_FOUND;
                } else if ((!errCode) && (strchr(pString, '~'))) {
                    errCode = LDRF_ERR_NOT_FOUND;
                }

                if ((!errCode) && (stringLength)) {
                    /* Save this new entry in string cache [0] */
                    pService->cachedString[0].pString = _strdup(pString);

                    if (!(pService->cachedString[0].pString)) {
                        errCode = LDRF_ERR_SYS;
                    } else {
                        pService->cachedString[0].pFileInfo
            = pService->cachedFile[0].pFileInfo;

                        pService->cachedString[0].index = stringIndex;
                    }
                } else {
                    pService->cachedString[0].pString = NULL;
                }

                if (pService->cachedString[0].pString) {
                    /* Put this string at end of active cache entries */
                    if (pService->numCachedStrings >= STRING_SERVICE_MAX_CACHED_STRINGS) {
                        // No inactive entry - zap a RANDOM active entry!
                        // Move LAST to zapped entry, makes room for new string
                        // as the LAST entry is now INACTIVE
                        TUShort zap;
                        pService->numCachedStrings = STRING_SERVICE_MAX_CACHED_STRINGS;
                        zap = (((TUShort)(rand())) % (STRING_SERVICE_MAX_CACHED_STRINGS - 1)) + 1;
                        index = STRING_SERVICE_MAX_CACHED_STRINGS;
                        FREE(pService->cachedString[zap].pString);
                        memcpy(&(pService->cachedString[zap]), &(pService->cachedString[index]), sizeof(pService->cachedString[0]));
                    } else {
                        index = ++(pService->numCachedStrings);
                    }
                    // Copy the "new" string's info (in [0]) to the
                    // inactive entry designated by [index].
                    memcpy(&(pService->cachedString[index]), &(pService->cachedString[0]), sizeof(pService->cachedString[0]));
                }
            } else {
                // String was already in cache - now reorganize the cache
                // for "hot file" access optimization
                memcpy(&(pService->cachedString[0]), &(pService->cachedString[index]), sizeof(pService->cachedString[0]));

                if (index > 1) {
                    // The optimization is to swap the "hit" entry with
                    // the previous one
                    --index;
                    memcpy(&(pService->cachedString[index + 1]), &(pService->cachedString[index]), sizeof(pService->cachedString[0]));
                    // Copy the "hit" file's info (in [0]) to the inactive entry
                    // designated by [index].
                    memcpy(&(pService->cachedString[index]), &(pService->cachedString[0]), sizeof(pService->cachedString[0]));
                }

                // Retrieve result from cache and put into pString
                stringLength = strlen(pService->cachedString[0].pString) + 1;

                if (stringLength > stringBufLength) {
                    strncpy(pString, pService->cachedString[0].pString, stringBufLength - 1);
                    pString[stringBufLength - 1] = '\0';
                    stringLength = stringBufLength;
                    errCode = LDRF_ERR_TRUNC;
                } else {
                    strcpy(pString, pService->cachedString[0].pString);
                }
            }
        }
    }

    if ((!errCode) && (pStringLength)) {
        * pStringLength = stringLength;
    }

    return errCode;
} /* end of: LdrfStringServiceRequestHelper */

LDRF_FN LdrfStringServiceRequest(PLdrfStringService pService, PUByteArray pProgID, TUByte scope, TULong index, LPSTR string, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_PARAM;
    PLdrfStringLocaleListElement pLocale;
    TUShort tempLength;

    if ((string) && (pLength) && (*pLength >= 2) && (pService) && (pProgID) && (scope <= 6) && (index > 0)) {
        errCode = LdrfValidateStringServicePointer(pService);

        if (!errCode) {
            errCode = LDRF_ERR_NOT_FOUND;
            pLocale = pService->pLocaleList;

            while (pLocale != NULL) {
                errCode = LdrfStringServiceRequestHelper(pService, pProgID, scope, pLocale->locale, index, *pLength, string, &tempLength);

                if (errCode != LDRF_ERR_NOT_FOUND) {
                    break;
                }
                pLocale = pLocale->pNextLocale;
            }

            if ((errCode == LDRF_ERR_NOT_FOUND) && (!(pService->localeListContainsENU))) {
                errCode = LdrfStringServiceRequestHelper(pService, pProgID, scope, ENU_LOCALE, index, *pLength, string, &tempLength);
            }

            if (errCode == LDRF_ERR_NOT_FOUND) {
/* Construct the "not available" message */
#define UNAVAIL_INDEX 1059

                TUByte progID[NUM_PROG_ID_BYTES];
                memset(progID, '\0', NUM_PROG_ID_BYTES);
                pLocale = pService->pLocaleList;

                while (pLocale != NULL) {
                    errCode = LdrfStringServiceRequestHelper(pService, pProgID, 0, pLocale->locale, UNAVAIL_INDEX, *pLength, string, &tempLength);

                    if (errCode != LDRF_ERR_NOT_FOUND) {
                        break;
                    }
                    pLocale = pLocale->pNextLocale;
                }

                if ((errCode == LDRF_ERR_NOT_FOUND) && (!(pService->localeListContainsENU))) {
                    errCode = LdrfStringServiceRequestHelper(pService, pProgID, 0, ENU_LOCALE, UNAVAIL_INDEX, *pLength, string, &tempLength);
                }

                if (errCode == LDRF_ERR_NOT_FOUND) {
                    if (*pLength >= 30) {
                        strcpy(string, "Message string is unavailable");
                        tempLength = 30;
                    } else if (*pLength >= 12) {
                        strcpy(string, "unavailable");
                        tempLength = 12;
                    } else {
                        strcpy(string, "?");
                        tempLength = 2;
                    }
                } else if (!errCode) {
                    errCode = LDRF_ERR_NOT_FOUND;
                }
            }

            if ((!errCode) || (errCode == LDRF_ERR_NOT_FOUND)) {
                * pLength = tempLength;
            }
        }
    }

    return errCode;
} /* end of: LdrfStringServiceRequest */
