/*
 * DRFTYPT.C - This is the Resource File API file responsible for access methods for type trees.
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif


/* SetCurrent is used after setting details in some nodes to move to next node */
/* It is also used before reading the next node type in a scan of existing tree */

static TLdrfErrCodes SetCurrent(PLdrfTypeTree pTypeTree) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

    if (pTypeTree->state == NVT_STATE_SET_DETAILS) {
        /* Tree creation ... */
        if (!pCurrent) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            /* At this point, we know we don't have a struct or a union or an array, since
             * those "set details" routines don't call SetCurrent. We make the parent node
             * the current node, then think about it some more. */
            PLdrfTypeNode pChild = pCurrent;
            pCurrent = pCurrent->pParent;

            while (pCurrent) {
                if ((pCurrent->nodeType == NVT_TYPE_STRUCT) || (pCurrent->nodeType == NVT_TYPE_UNION)) {
                    /* Update the typeSize of the aggregate */
                    if (pCurrent->nodeType == NVT_TYPE_STRUCT) {
                        TUByte bitsUsed = 0;

                        if (pChild->nodeType == NVT_TYPE_BITF) {
                            bitsUsed = pChild->u.NVTBitf.bitfOffset + pChild->u.NVTBitf.bitfSize;

                            if (bitsUsed >= 8) {
                                bitsUsed = 0;
                            }
                        }
                        pCurrent->u.NVTStructUnion.bitfBitsUsed = bitsUsed;
                        pCurrent->typeSize += pChild->typeSize;
                    } else if (pCurrent->nodeType == NVT_TYPE_UNION) {
                        if (pCurrent->typeSize < pChild->typeSize) {
                            pCurrent->typeSize = pChild->typeSize;
                        }
                    }
                    /* If current node is an incomplete struct, stop here ... */
                    if (pCurrent->u.NVTStructUnion.currentCount < pCurrent->u.NVTStructUnion.numFields) {
                        break;
                    }
                } else if (pCurrent->nodeType == NVT_TYPE_ARRAY) {
                    /* Calculate the typeSize of the array */
                    pCurrent->typeSize = pChild->typeSize * pCurrent->u.NVTArray.numElements;
                }
                /* Keep moving up the parent node chain */
                pChild = pCurrent;
                pCurrent = pCurrent->pParent;
            }

            if (pCurrent) {
                pTypeTree->state = NVT_STATE_PARTIAL; /* Incomplete struct */
                errCode = LDRF_ERR_INCOMPLETE;
            } else {
                /* The tree is complete - no more nodes after this.  Set to neutral state. */
                pTypeTree->state = NVT_STATE_NONE;
            }
        }
    } else if (pTypeTree->state == NVT_STATE_SCAN_FIND) {
        if (!pCurrent) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            pTypeTree->state = NVT_STATE_SCAN;
        }
    } else if (pTypeTree->state == NVT_STATE_SCAN) {
        /* Tree scanning ... */
        if (!pCurrent) {
            pCurrent = pTypeTree->pNodes;
        } else if (pCurrent->nodeType == NVT_TYPE_ARRAY) {
            ++(pCurrent->u.NVTArray.currentCount);
            pCurrent = pCurrent->u.NVTArray.pElement;
        } else if ((pCurrent->nodeType == NVT_TYPE_STRUCT) || (pCurrent->nodeType == NVT_TYPE_UNION)) {
            pCurrent = pCurrent->u.NVTStructUnion.pFields;
        } else {
            while (pCurrent) {
                if (pCurrent->pFieldLink) {
                    /* Just go to the next field */
                    pCurrent = pCurrent->pFieldLink;
                    break;
                } else {
                    /* There is no next field - go to parent */
                    pCurrent = pCurrent->pParent;

                    if ((pCurrent) && (pCurrent->nodeType == NVT_TYPE_ARRAY)) {
                        if (pCurrent->u.NVTArray.currentCount < pCurrent->u.NVTArray.numElements) {
                            /* Repeated traversal of element node */
                            ++(pCurrent->u.NVTArray.currentCount);
                            pCurrent = pCurrent->u.NVTArray.pElement;
                            pTypeTree->arrayTypeOffsetAdder += pCurrent->typeSize;
                            break;
                        } else {
                            pTypeTree->arrayTypeOffsetAdder -= ((pCurrent->u.NVTArray.numElements - 1) * pCurrent->u.NVTArray.pElement->typeSize);
                        }
                    }
                }
            }
        }

        if (!pCurrent) {
            pTypeTree->state = NVT_STATE_NONE;
            errCode = LDRF_ERR_NO_MORE_NODES;
        }
    } else {
        errCode = LDRF_ERR_SEQUENCE;
    }

    pTypeTree->pCurrent = pCurrent;

    return errCode;
} /* end of: SetCurrent */


/* FreeTypeNodes is a recursive routine that gobbles up & frees all type nodes of a tree. */
static void LdrfFreeTypeNodes(PLdrfTypeNode pNodes) {
    while (pNodes) {
        PLdrfTypeNode pTemp = pNodes;
        FREE(pTemp->name);

        if (pTemp->nodeType == NVT_TYPE_ARRAY) {
            LdrfFreeTypeNodes(pTemp->u.NVTArray.pElement);
        } else if ((pTemp->nodeType == NVT_TYPE_STRUCT) || (pTemp->nodeType == NVT_TYPE_UNION)) {
            LdrfFreeTypeNodes(pTemp->u.NVTStructUnion.pFields);
        }
        /* Point to next node in field link chain, then free the current node */
        pNodes = pTemp->pFieldLink;
        FREE(pTemp);
    }
} /* end of: LdrfFreeTypeNodes */

LDRF_FN LdrfFreeTypeTree(PLdrfTypeTree pTypeTree) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        LdrfFreeTypeNodes(pTypeTree->pNodes);
        LdrfFreeTypeTreePtr(&pTypeTree);
    }

    return errCode;
} /* end of: LdrfFreeTypeTree */

LDRF_FN LdrfNewTypeTreeNode(TNVTType newNodeType, LPCSTR name, TUByte resNmScope, TULong resNmIndex, TUByte resCmtScope, TULong resCmtIndex, TUByte resUntScope,
    TULong resUntIndex, PLdrfTypeTree* ppTypeTree)
{
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfTypeTree pTypeTree;

    if (!ppTypeTree) {
        errCode = LDRF_ERR_PARAM;
    } else {
        pTypeTree = *ppTypeTree;

        if (!pTypeTree) {
            /* This is a new type tree */
            errCode = LdrfNewTypeTree(&pTypeTree);
            * ppTypeTree = pTypeTree;
        } else {
            /* Adding a node to an existing tree */
            errCode = LdrfValidateTypeTreePointer(pTypeTree);
        }

        if (!errCode) {
            if ((pTypeTree->state != NVT_STATE_NEW) && (pTypeTree->state != NVT_STATE_PARTIAL)) {
                errCode = LDRF_ERR_SEQUENCE;
            } else {
                PLdrfTypeNode pParent = pTypeTree->pCurrent;
                PLdrfTypeNode pNewNode = NEW(TLdrfTypeNode);
                LPSTR pNameCopy = _strdup(name);

                if ((!pNewNode) || (!pNameCopy)) {
                    errCode = LDRF_ERR_SYS;
                } else if ((newNodeType == NVT_TYPE_ARRAY) && (!pParent)) {
                    errCode = LDRF_ERR_PARAM;
                } else if ((newNodeType == NVT_TYPE_BITF) && ((!pParent) || (pParent->nodeType != NVT_TYPE_STRUCT))) {
                    errCode = LDRF_ERR_PARAM;
                } else {
                    if (newNodeType == NVT_TYPE_DOUBLE_FLOAT) {
                        pTypeTree->containsDataTypeDoubleFloat = TRUE;
                    } else if (newNodeType == NVT_TYPE_UNSIGNED_QUAD) {
                        pTypeTree->containsDataTypeUnsignedQuad = TRUE;
                    } else if (newNodeType == NVT_TYPE_UNSIGNED_INT64) {
                        pTypeTree->containsDataTypeUnsignedInt64 = TRUE;
                    } else if (newNodeType == NVT_TYPE_SIGNED_INT64) {
                        pTypeTree->containsDataTypeSignedInt64 = TRUE;
                    }
                    pNewNode->nodeType = newNodeType;
                    pNewNode->pParent = pParent;
                    pNewNode->pFieldLink = NULL;
                    pNewNode->name = pNameCopy;
                    pNewNode->resNmScope = resNmScope;
                    pNewNode->resNmIndex = resNmIndex;
                    pNewNode->resCmtScope = resCmtScope;
                    pNewNode->resCmtIndex = resCmtIndex;
                    pNewNode->resUntScope = resUntScope;
                    pNewNode->resUntIndex = resUntIndex;
                    pNewNode->typeSize = 0;

                    if (!pParent) {
                        pNewNode->typeOffset = 0;
                    } else {
                        /* Get the type offset from the parent */
                        pNewNode->typeOffset = pParent->typeOffset;

                        if (pParent->nodeType == NVT_TYPE_STRUCT) {
                            /* Add in the size of previous fields */
                            pNewNode->typeOffset += pParent->typeSize;
                        }
                    }

                    if (((NVT_TYPE_SIGNED_CHAR <= pNewNode->nodeType) && (pNewNode->nodeType <= NVT_TYPE_ENUM)) || (pNewNode->nodeType == NVT_TYPE_SIGNED_QUAD)
                        || (pNewNode->nodeType == NVT_TYPE_UNSIGNED_QUAD)) {
                        pNewNode->u.NVTScalar.invalidValuePresent = F;
                    } else if ((pNewNode->nodeType == NVT_TYPE_SIGNED_INT64) || pNewNode->nodeType == NVT_TYPE_UNSIGNED_INT64) {                        
                        pNewNode->u.NVTScalar64.invalidValuePresent = F;
                    } else if (pNewNode->nodeType == NVT_TYPE_REFERENCE) {
                        ++(pTypeTree->refCounter);
                        pNewNode->u.NVTReference.pBackLink = pParent;
                    } else if (pNewNode->nodeType == NVT_TYPE_ARRAY) {
                        pNewNode->u.NVTArray.pElement = NULL;
                    } else if ((pNewNode->nodeType == NVT_TYPE_STRUCT) || (pNewNode->nodeType == NVT_TYPE_UNION)) {
                        pNewNode->u.NVTStructUnion.pFields = NULL;
                    }

                    if (pTypeTree->state == NVT_STATE_NEW) {
                        pTypeTree->pNodes = pNewNode;
                    } else {
                        if (pParent->nodeType == NVT_TYPE_ARRAY) {
                            pParent->u.NVTArray.pElement = pNewNode;
                        } else if ((pParent->nodeType == NVT_TYPE_STRUCT) || (pParent->nodeType == NVT_TYPE_UNION)) {
                            if (!(pParent->u.NVTStructUnion.pCurrent)) {
                                /* The first field in the struct or union */
                                pParent->u.NVTStructUnion.pFields = pNewNode;
                            } else {
                                pParent->u.NVTStructUnion.pCurrent->pFieldLink = pNewNode;
                                /* Override the backlink setting to be the previous field */
                                if (pNewNode->nodeType == NVT_TYPE_REFERENCE) {
                                    pNewNode->u.NVTReference.pBackLink = pParent->u.NVTStructUnion.pCurrent;
                                }
                            }
                            /* This remembers the current field at this level */
                            pParent->u.NVTStructUnion.pCurrent = pNewNode;
                            ++(pParent->u.NVTStructUnion.currentCount);
                        } else {
                            errCode = LDRF_ERR_TYPE_TREE;
                        }
                    }

                    if (!errCode) {
                        pTypeTree->pCurrent = pNewNode;
                        pTypeTree->state = NVT_STATE_SET_DETAILS;
                        errCode = LDRF_ERR_INCOMPLETE;
                    }
                }

                if (errCode != LDRF_ERR_INCOMPLETE) {
                    if (pNewNode) {
                        FREE(pNewNode);
                    }

                    if (pNameCopy) {
                        FREE(pNameCopy);
                    }
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfNewTypeTreeNode */

LDRF_FN LdrfSetScalarDetails(PLdrfTypeTree pTypeTree, TLong minValid, TLong maxValid, TShort scaleA, TShort scaleB, TShort scaleC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType < NVT_TYPE_SIGNED_CHAR)
            || ((pCurrent->nodeType > NVT_TYPE_UNSIGNED_LONG) && (pCurrent->nodeType != NVT_TYPE_SIGNED_QUAD)
                && (pCurrent->nodeType != NVT_TYPE_UNSIGNED_QUAD)))
            {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTScalar.minValid = minValid;
            pCurrent->u.NVTScalar.maxValid = maxValid;
            pCurrent->u.NVTScalar.scaleA = scaleA;
            pCurrent->u.NVTScalar.scaleB = scaleB;
            pCurrent->u.NVTScalar.scaleC = scaleC;

            switch (pCurrent->nodeType) {
                case NVT_TYPE_SIGNED_CHAR:
                case NVT_TYPE_UNSIGNED_CHAR:
                case NVT_TYPE_SIGNED_SHORT:
                case NVT_TYPE_UNSIGNED_SHORT:
                    pCurrent->typeSize = 1;
                    break;

                case NVT_TYPE_SIGNED_LONG:
                case NVT_TYPE_UNSIGNED_LONG:
                    pCurrent->typeSize = 2;
                    break;

                case NVT_TYPE_SIGNED_QUAD:
                case NVT_TYPE_UNSIGNED_QUAD:
                    pCurrent->typeSize = 4;
                    break;

                default:
                    /* Adding this clause here prevents some compilers
                     * from creating warning messages */
                    ;
            }
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetScalarDetails */

LDRF_FN LdrfSetScalar64Details  ( PLdrfTypeTree pTypeTree,
				  __DRF_INT64 minValid, __DRF_INT64 maxValid,
                  TShort scaleA, TShort scaleB, TShort scaleC ) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || 
             (pCurrent->nodeType != NVT_TYPE_UNSIGNED_INT64 && 
              pCurrent->nodeType != NVT_TYPE_SIGNED_INT64)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTScalar64.minValid = minValid;
            pCurrent->u.NVTScalar64.maxValid = maxValid;
            pCurrent->u.NVTScalar64.scaleA = scaleA;
            pCurrent->u.NVTScalar64.scaleB = scaleB;
            pCurrent->u.NVTScalar64.scaleC = scaleC;
            pCurrent->typeSize = 8;
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetScalar64Details */


LDRF_FN LdrfSetScalarInvalidValue(PLdrfTypeTree pTypeTree, TLong invalidValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType < NVT_TYPE_SIGNED_CHAR)
            || ((pCurrent->nodeType > NVT_TYPE_UNSIGNED_LONG) && (pCurrent->nodeType != NVT_TYPE_SIGNED_QUAD)
                && (pCurrent->nodeType != NVT_TYPE_UNSIGNED_QUAD)))
            {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTScalar.invalidValuePresent = T;
            pCurrent->u.NVTScalar.invalidVal = invalidValue;
        }
    }

    return errCode;
} /* end of: LdrfSetScalarInvalidValue */

LDRF_FN LdrfSetScalar64InvalidValue(PLdrfTypeTree pTypeTree, __DRF_INT64 invalidValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || 
             (pCurrent->nodeType != NVT_TYPE_SIGNED_INT64 &&
              pCurrent->nodeType != NVT_TYPE_UNSIGNED_INT64)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTScalar64.invalidValuePresent = T;
            pCurrent->u.NVTScalar64.invalidVal = invalidValue;
        }
    }

    return errCode;
} /* end of: LdrfSetScalar64InvalidValue */


LDRF_FN LdrfSetFloatDetails(PLdrfTypeTree pTypeTree, float minValid, float maxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_FLOAT)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTFloat.minValid = minValid;
            pCurrent->u.NVTFloat.maxValid = maxValid;
            pCurrent->typeSize = 4;
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetFloatDetails */

LDRF_FN LdrfSetDoubleFloatDetails(PLdrfTypeTree pTypeTree, double minValid, double maxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_DOUBLE_FLOAT)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTDouble.minValid = minValid;
            pCurrent->u.NVTDouble.maxValid = maxValid;
            pCurrent->typeSize = 8;
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetDoubleFloatDetails */

LDRF_FN LdrfSetBitfieldDetails(PLdrfTypeTree pTypeTree, TUByte bitfSize, TUByte bitfOffset, TBool bitfSigned, TLong minValid, TLong maxValid, TShort scaleA,
    TShort scaleB, TShort scaleC)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_BITF)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else if (bitfOffset + bitfSize > 8) {
            errCode = LDRF_ERR_PARAM;
        } else {
            pCurrent->u.NVTBitf.bitfSize = bitfSize;
            pCurrent->u.NVTBitf.bitfOffset = bitfOffset;
            pCurrent->u.NVTBitf.bitfSigned = bitfSigned;
            pCurrent->u.NVTBitf.minValid = minValid;
            pCurrent->u.NVTBitf.maxValid = maxValid;
            pCurrent->u.NVTBitf.scaleA = scaleA;
            pCurrent->u.NVTBitf.scaleB = scaleB;
            pCurrent->u.NVTBitf.scaleC = scaleC;

            if ((pCurrent->pParent) && (pCurrent->pParent->nodeType == NVT_TYPE_STRUCT)) {
                TUShort bitsUsed = pCurrent->pParent->u.NVTStructUnion.bitfBitsUsed;

                if ((bitsUsed == 0) || (bitsUsed > bitfOffset)) {
                    pCurrent->typeSize = 1;
                // } else if (bitsUsed > bitfOffset) {
                // pCurrent->typeSize = 1;
                // ++(pCurrent->typeOffset); EPR 10587
                } else {
                    --(pCurrent->typeOffset);
                }
                errCode = SetCurrent(pTypeTree);
            } else {
                errCode = LDRF_ERR_INTERNAL;
            }
        }
    }

    return errCode;
} /* end of: LdrfSetBitfieldDetails */

LDRF_FN LdrfSetEnumDetails(PLdrfTypeTree pTypeTree, TUByte enumScope, TUShort enumIndex, TLong minValid, TLong maxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_ENUM)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTEnum.enumScope = enumScope;
            pCurrent->u.NVTEnum.enumIndex = enumIndex;
            pCurrent->u.NVTEnum.minValid = minValid;
            pCurrent->u.NVTEnum.maxValid = maxValid;
            pCurrent->typeSize = 1;
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetEnumDetails */

LDRF_FN LdrfSetArrayDetails(PLdrfTypeTree pTypeTree, TUShort numElements) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_ARRAY)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTArray.numElements = numElements;
            pCurrent->u.NVTArray.pElement = NULL;
            errCode = LDRF_ERR_INCOMPLETE;
            pTypeTree->state = NVT_STATE_PARTIAL;
        }
    }

    return errCode;
} /* end of: LdrfSetArrayDetails */

LDRF_FN LdrfSetStructUnionDetails(PLdrfTypeTree pTypeTree, TUShort numFields) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || ((pCurrent->nodeType != NVT_TYPE_STRUCT) && (pCurrent->nodeType != NVT_TYPE_UNION))) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTStructUnion.numFields = numFields;
            pCurrent->u.NVTStructUnion.currentCount = 0;
            pCurrent->u.NVTStructUnion.pFields = NULL;
            pCurrent->u.NVTStructUnion.pCurrent = NULL;
            pCurrent->u.NVTStructUnion.bitfBitsUsed = 0;
            errCode = LDRF_ERR_INCOMPLETE;
            pTypeTree->state = NVT_STATE_PARTIAL;
        }
    }

    return errCode;
} /* end of: LdrfSetStructUnionDetails */

LDRF_FN LdrfSetReferenceDetails(PLdrfTypeTree pTypeTree, TUByte typeScope, TUShort typeIndex, TUShort typeSize) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SET_DETAILS) || (pCurrent->nodeType != NVT_TYPE_REFERENCE)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pCurrent->u.NVTReference.scope = typeScope;
            pCurrent->u.NVTReference.index = typeIndex;
            pCurrent->typeSize = typeSize;
            errCode = SetCurrent(pTypeTree);
        }
    }

    return errCode;
} /* end of: LdrfSetReferenceDetails */

LDRF_FN LdrfScanTypeTree(PLdrfTypeTree pTypeTree, PUShort pTypeSize, PBool pHasRefs) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        if ((pTypeTree->state != NVT_STATE_NONE) && (pTypeTree->state != NVT_STATE_SCAN) && (pTypeTree->state != NVT_STATE_SCAN_FIND)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            pTypeTree->pCurrent = NULL;
            pTypeTree->state = NVT_STATE_SCAN;
            pTypeTree->arrayTypeOffsetAdder = 0;

            if (pTypeSize) {
                if (pTypeTree->pNodes) {
                    * pTypeSize = pTypeTree->pNodes->typeSize;
                } else {
                    * pTypeSize = 0;
                }
            }

            if (pHasRefs) {
                if (pTypeTree->refCounter) {
                    * pHasRefs = T;
                } else {
                    * pHasRefs = F;
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfScanTypeTree */

LDRF_FN LdrfFindTypeTreeNode(PLdrfTypeTree pTypeTree, TBool relative, LPCSTR pFieldName) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        if ((pTypeTree->state != NVT_STATE_NONE) && (pTypeTree->state != NVT_STATE_SCAN)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            PLdrfTypeNode pCurrent;

            if (relative) {
                pCurrent = pTypeTree->pCurrent;

                if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || ((pCurrent->nodeType != NVT_TYPE_STRUCT) && (pCurrent->nodeType != NVT_TYPE_UNION)))
                    {
                    errCode = LDRF_ERR_SEQUENCE;
                }
            } else {
                pCurrent = pTypeTree->pNodes;
                pTypeTree->arrayTypeOffsetAdder = 0;
            }

            if (!errCode) {
                char name[MAX_NC_NAME + 1];
                const char* pFld, * pEnd1, * pEnd2;
                TUShort len;
                TBool arrayFlag;
                pFld = pFieldName;

                while ((pFld) && (pCurrent) && (!errCode)) {
                    /* Get the next part of field name into "name" */
                    if (*pFld == EOS) {
                        break;
                    }
                    arrayFlag = F;
                    pEnd1 = strchr(pFld, '.');
                    pEnd2 = strchr(pFld, '[');

                    if (pEnd2) {
                        if ((!pEnd1) || (pEnd1 > pEnd2)) {
                            arrayFlag = T;
                            pEnd1 = pEnd2;
                        }
                    } else if (!pEnd1) {
                        pEnd1 = strchr(pFld, EOS);
                    }
                    len = pEnd1 - pFld;

                    if (len > MAX_NC_NAME) {
                        len = MAX_NC_NAME;
                    }
                    strncpy(name, pFld, len);
                    name[len] = EOS;
                    pFld = pEnd1;

                    if (*pFld != EOS) {
                        ++pFld; /* Skip past the '[' or '.' separator */
                    }
                    /* Look for the name in a structure */
                    if ((pCurrent->nodeType == NVT_TYPE_STRUCT) || (pCurrent->nodeType == NVT_TYPE_UNION)) {
                        pCurrent = pCurrent->u.NVTStructUnion.pFields;

                        while (pCurrent) {
                            if (!strcmp(pCurrent->name, name)) {
                                break;
                            }
                            pCurrent = pCurrent->pFieldLink;
                        }
                    } else {
                        pCurrent = NULL;
                    }

                    if ((arrayFlag) && (pCurrent) && (pCurrent->nodeType == NVT_TYPE_ARRAY)) {
                        /* This reads the array index (integer) */
                        if (isdigit((int)(* pFld))) {
                            int index;
                            sscanf(pFld, "%d", &index);

                            if (index >= pCurrent->u.NVTArray.numElements) {
                                errCode = LDRF_ERR_PARAM;
                            } else {
                                pCurrent->u.NVTArray.currentCount = index;

                                while ((* pFld) && (isdigit((int)(* pFld)))) {
                                    ++pFld;
                                }

                                if (*pFld != ']') {
                                    errCode = LDRF_ERR_PARAM;
                                } else {
                                    ++pFld;
                                    pCurrent = pCurrent->u.NVTArray.pElement;

                                    if (*pFld == '.') {
                                        ++pFld; /* ']' was followed by '.' */
                                    }
                                }
                            }

                            if (pCurrent) {
                                pTypeTree->arrayTypeOffsetAdder += (pCurrent->typeSize * index);
                            }
                        } else if (*pFld == ']') {
                            pCurrent->u.NVTArray.currentCount = 0;
                            pCurrent = pCurrent->u.NVTArray.pElement;
                            ++pFld;
                        } else if (*pFld) {
                            errCode = LDRF_ERR_PARAM;
                        }
                    }
                } /* end of: while ((pFld) && (pCurrent) && (!errCode)) */

                if (!pCurrent) {
                    errCode = LDRF_ERR_NOT_FOUND;
                } else {
                    pTypeTree->pCurrent = pCurrent;
                    pTypeTree->state = NVT_STATE_SCAN_FIND;
                }
            }
        }

        if (errCode) {
            pTypeTree->state = NVT_STATE_NONE;
        }
    }

    return errCode;
} /* end of: LdrfFindTypeTreeNode */

LDRF_FN LdrfReadTypeTreeNode(PLdrfTypeTree pTypeTree, PNVTType pNodeType, PUShort pTypeOffset, PUShort pTypeSize, LPSTR pName, PUShort pLength,
    PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pResUntScope, PULong pResUntIndex)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        errCode = SetCurrent(pTypeTree);

        if (!errCode) {
            PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

            if (pNodeType) {
                * pNodeType = pCurrent->nodeType;
            }

            if (pTypeOffset) {
                TUShort typeOffset = pCurrent->typeOffset;
                typeOffset += pTypeTree->arrayTypeOffsetAdder;
                * pTypeOffset = typeOffset;
            }

            if (pTypeSize) {
                * pTypeSize = pCurrent->typeSize;
            }

            if ((*(pCurrent->name) == EOS) && (pName) && (pCurrent->pParent) && (pCurrent->pParent->nodeType == NVT_TYPE_ARRAY)) {
                char tempName[MAX_NC_NAME + 10];
                char buf[10];
                strncpy(tempName, pCurrent->pParent->name, MAX_NC_NAME);
                tempName[MAX_NC_NAME + 1] = EOS;
                buf[0] = '[';
                _itoa(pCurrent->pParent->u.NVTArray.currentCount - 1, buf + 1, 10);
                strcat(buf, "]");
                strncat(tempName, buf, 9);
                tempName[MAX_NC_NAME + 9] = EOS;
                errCode = LdrfTransferStringOut(pName, pLength, tempName);
            } else {
                errCode = LdrfTransferStringOut(pName, pLength, pCurrent->name);
            }

            if (pResNmScope) {
                * pResNmScope = pCurrent->resNmScope;
            }

            if (pResNmIndex) {
                * pResNmIndex = pCurrent->resNmIndex;
            }

            if (pResCmtScope) {
                * pResCmtScope = pCurrent->resCmtScope;
            }

            if (pResCmtIndex) {
                * pResCmtIndex = pCurrent->resCmtIndex;
            }

            if (pResUntScope) {
                * pResUntScope = pCurrent->resUntScope;
            }

            if (pResUntIndex) {
                * pResUntIndex = pCurrent->resUntIndex;
            }

            if (!errCode) {
                errCode = LDRF_ERR_INCOMPLETE;
            }
        }
    }

    return errCode;
} /* end of: LdrfReadTypeTreeNode */

LDRF_FN LdrfGetScalarDetails(PLdrfTypeTree pTypeTree, PLong pMinValid, PLong pMaxValid, PShort pScaleA, PShort pScaleB, PShort pScaleC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType < NVT_TYPE_SIGNED_CHAR)
            || ((pCurrent->nodeType > NVT_TYPE_UNSIGNED_LONG) && (pCurrent->nodeType != NVT_TYPE_SIGNED_QUAD)
                && (pCurrent->nodeType != NVT_TYPE_UNSIGNED_QUAD)))
            {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTScalar.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTScalar.maxValid;
            }

            if (pScaleA) {
                * pScaleA = pCurrent->u.NVTScalar.scaleA;
            }

            if (pScaleB) {
                * pScaleB = pCurrent->u.NVTScalar.scaleB;
            }

            if (pScaleC) {
                * pScaleC = pCurrent->u.NVTScalar.scaleC;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetScalarDetails */

LDRF_FN LdrfGetScalar64Details	( PLdrfTypeTree pTypeTree,
				  __DRF_PINT64 pMinValid, __DRF_PINT64 pMaxValid,
                  PShort pScaleA, PShort pScaleB, PShort pScaleC ) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || 
             (pCurrent->nodeType != NVT_TYPE_SIGNED_INT64 && 
              pCurrent->nodeType != NVT_TYPE_UNSIGNED_INT64)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTScalar64.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTScalar64.maxValid;
            }

            if (pScaleA) {
                * pScaleA = pCurrent->u.NVTScalar64.scaleA;
            }

            if (pScaleB) {
                * pScaleB = pCurrent->u.NVTScalar64.scaleB;
            }

            if (pScaleC) {
                * pScaleC = pCurrent->u.NVTScalar64.scaleC;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetScalarDetails */

LDRF_FN LdrfGetScalarInvalidValue(PLdrfTypeTree pTypeTree, PBoolByte pInvalidValuePresent, PLong pInvalidValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType < NVT_TYPE_SIGNED_CHAR)
            || ((pCurrent->nodeType > NVT_TYPE_UNSIGNED_LONG) && (pCurrent->nodeType != NVT_TYPE_SIGNED_QUAD)
                && (pCurrent->nodeType != NVT_TYPE_UNSIGNED_QUAD)))
            {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pInvalidValuePresent) {
                * pInvalidValuePresent = pCurrent->u.NVTScalar.invalidValuePresent;

                if ((* pInvalidValuePresent) && (pInvalidValue)) {
                    * pInvalidValue = pCurrent->u.NVTScalar.invalidVal;
                }
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetScalarInvalidValue */

LDRF_FN LdrfGetScalar64InvalidValue(PLdrfTypeTree pTypeTree, PBoolByte pInvalidValuePresent, __DRF_PINT64 pInvalidValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || 
             (pCurrent->nodeType != NVT_TYPE_UNSIGNED_INT64 &&
              pCurrent->nodeType != NVT_TYPE_SIGNED_INT64)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pInvalidValuePresent) {
                * pInvalidValuePresent = pCurrent->u.NVTScalar64.invalidValuePresent;

                if ((* pInvalidValuePresent) && (pInvalidValue)) {
                    * pInvalidValue = pCurrent->u.NVTScalar64.invalidVal;
                }
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetScalar64InvalidValue */

LDRF_FN LdrfGetFloatDetails(PLdrfTypeTree pTypeTree, float* pMinValid, float* pMaxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_FLOAT)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTFloat.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTFloat.maxValid;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetFloatDetails */

LDRF_FN LdrfGetDoubleFloatDetails(PLdrfTypeTree pTypeTree, double* pMinValid, double* pMaxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_DOUBLE_FLOAT)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTDouble.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTDouble.maxValid;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetDoubleFloatDetails */

LDRF_FN LdrfGetBitfieldDetails(PLdrfTypeTree pTypeTree, PUByte pBitfSize, PUByte pBitfOffset, PBool pBitfSigned, PLong pMinValid, PLong pMaxValid,
    PShort pScaleA, PShort pScaleB, PShort pScaleC)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_BITF)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pBitfSize) {
                * pBitfSize = pCurrent->u.NVTBitf.bitfSize;
            }

            if (pBitfOffset) {
                * pBitfOffset = pCurrent->u.NVTBitf.bitfOffset;
            }

            if (pBitfSigned) {
                * pBitfSigned = pCurrent->u.NVTBitf.bitfSigned;
            }

            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTBitf.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTBitf.maxValid;
            }

            if (pScaleA) {
                * pScaleA = pCurrent->u.NVTBitf.scaleA;
            }

            if (pScaleB) {
                * pScaleB = pCurrent->u.NVTBitf.scaleB;
            }

            if (pScaleC) {
                * pScaleC = pCurrent->u.NVTBitf.scaleC;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetBitfieldDetails */

LDRF_FN LdrfGetEnumDetails(PLdrfTypeTree pTypeTree, PUByte pEnumScope, PUShort pEnumIndex, PLong pMinValid, PLong pMaxValid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_ENUM)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pEnumScope) {
                * pEnumScope = pCurrent->u.NVTEnum.enumScope;
            }

            if (pEnumIndex) {
                * pEnumIndex = pCurrent->u.NVTEnum.enumIndex;
            }

            if (pMinValid) {
                * pMinValid = pCurrent->u.NVTEnum.minValid;
            }

            if (pMaxValid) {
                * pMaxValid = pCurrent->u.NVTEnum.maxValid;
            }

            if (!errCode) {
                errCode = LDRF_ERR_INCOMPLETE;
            }
        }
    }

    return errCode;
} /* end of: LdrfGetEnumDetails */

LDRF_FN LdrfGetArrayDetails(PLdrfTypeTree pTypeTree, TBool multiple, PUShort pNumElements) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pTypeTree->pCurrent->nodeType != NVT_TYPE_ARRAY)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (multiple) {
                pCurrent->u.NVTArray.currentCount = 0;
            } else {
                pCurrent->u.NVTArray.currentCount = pCurrent->u.NVTArray.numElements - 1;
                pTypeTree->arrayTypeOffsetAdder += (pCurrent->u.NVTArray.currentCount * pCurrent->u.NVTArray.pElement->typeSize);
            }

            if (pNumElements) {
                * pNumElements = pCurrent->u.NVTArray.numElements;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetArrayDetails */

LDRF_FN LdrfGetStructUnionDetails(PLdrfTypeTree pTypeTree, PUShort pNumFields) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || ((pCurrent->nodeType != NVT_TYPE_STRUCT) && (pCurrent->nodeType != NVT_TYPE_UNION))) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pNumFields) {
                * pNumFields = pCurrent->u.NVTStructUnion.numFields;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetStructUnionDetails */

LDRF_FN LdrfGetReferenceDetails(PLdrfTypeTree pTypeTree, PUByte pTypeScope, PUShort pTypeIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        PLdrfTypeNode pCurrent = pTypeTree->pCurrent;

        if ((pTypeTree->state != NVT_STATE_SCAN) || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_REFERENCE)) {
            errCode = LDRF_ERR_SEQUENCE;
        } else {
            if (pTypeScope) {
                * pTypeScope = pCurrent->u.NVTReference.scope;
            }

            if (pTypeIndex) {
                * pTypeIndex = pCurrent->u.NVTReference.index;
            }
            errCode = LDRF_ERR_INCOMPLETE;
        }
    }

    return errCode;
} /* end of: LdrfGetReferenceDetails */

LDRF_FN LdrfGraftReference(PLdrfTypeTree pTypeTree, PLdrfTypeTree pReferentTree) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);

    if (!errCode) {
        errCode = LdrfValidateTypeTreePointer(pReferentTree);

        if (!errCode) {
            PLdrfTypeNode pCurrent = pTypeTree->pCurrent;
            PLdrfTypeNode pNewNodes = pReferentTree->pNodes;

            if ((pTypeTree->state != NVT_STATE_SCAN) || ((pReferentTree->state != NVT_STATE_NONE) && (pReferentTree->state != NVT_STATE_SCAN)) || (!pNewNodes)
                || (!pCurrent) || (pCurrent->nodeType != NVT_TYPE_REFERENCE))
                {
                errCode = LDRF_ERR_SEQUENCE;
            } else if (!(pTypeTree->refCounter)) {
                errCode = LDRF_ERR_INTERNAL;
            } else {
                PLdrfTypeNode pBackLink = pCurrent->u.NVTReference.pBackLink;

                if (!pBackLink) {
                    if (pTypeTree->pNodes == pCurrent) {
                        pTypeTree->pNodes = pNewNodes;
                    } else {
                        errCode = LDRF_ERR_INTERNAL;
                    }
                } else if ((pBackLink->nodeType == NVT_TYPE_ARRAY) && (pBackLink->u.NVTArray.pElement == pCurrent)) {
                    pBackLink->u.NVTArray.pElement = pNewNodes;
                } else if (((pBackLink->nodeType == NVT_TYPE_STRUCT) || (pBackLink->nodeType == NVT_TYPE_UNION))
                    && (pBackLink->u.NVTStructUnion.pFields == pCurrent))
                    {
                    pBackLink->u.NVTStructUnion.pFields = pNewNodes;
                } else {
                    pBackLink->pFieldLink = pNewNodes;
                }

                if (!errCode) {
                    /* The piece of tree being grafted in must have its
                     * offsets updated to reflect its position in the
                     * larger tree. */
                    PLdrfTypeNode pNewScan = pNewNodes;

                    while (pNewScan) {
                        pNewScan->typeOffset += pCurrent->typeOffset;

                        if (pNewScan->nodeType == NVT_TYPE_ARRAY) {
                            pNewScan = pNewScan->u.NVTArray.pElement;
                        } else if ((pNewScan->nodeType == NVT_TYPE_STRUCT) || (pNewScan->nodeType == NVT_TYPE_UNION)) {
                            pNewScan = pNewScan->u.NVTStructUnion.pFields;
                        } else {
                            while (pNewScan) {
                                if (pNewScan->pFieldLink) {
                                    /* Just go to the next field */
                                    pNewScan = pNewScan->pFieldLink;
                                    break;
                                } else {
                                    /* There is no next field - go to parent */
                                    pNewScan = pNewScan->pParent;
                                }
                            }
                        }
                    }
                    /* Update tree control information */
                    pTypeTree->refCounter -= 1;
                    pTypeTree->refCounter += pReferentTree->refCounter;
                    /* Transfer links & info from old ref node */
                    pNewNodes->pParent = pCurrent->pParent;
                    pNewNodes->pFieldLink = pCurrent->pFieldLink;
                    pCurrent->pFieldLink = NULL;

                    if (pNewNodes->nodeType == NVT_TYPE_REFERENCE) {
                        pNewNodes->u.NVTReference.pBackLink = pBackLink;
                    }

                    if ((pNewNodes->pFieldLink)
                        && ((pNewNodes->pFieldLink->nodeType == NVT_TYPE_REFERENCE) && (pNewNodes->pFieldLink->u.NVTReference.pBackLink == pCurrent)))
                        {
                        pNewNodes->pFieldLink->u.NVTReference.pBackLink = pNewNodes;
                    }

                    if (pCurrent->resNmIndex) {
                        pNewNodes->resNmScope = pCurrent->resNmScope;
                        pNewNodes->resNmIndex = pCurrent->resNmIndex;
                    }

                    if (pCurrent->resCmtIndex) {
                        pNewNodes->resCmtScope = pCurrent->resCmtScope;
                        pNewNodes->resCmtIndex = pCurrent->resCmtIndex;
                    }

                    if (pCurrent->resUntIndex) {
                        pNewNodes->resUntScope = pCurrent->resUntScope;
                        pNewNodes->resUntIndex = pCurrent->resUntIndex;
                    }

                    if (pCurrent->name) {
                        FREE(pNewNodes->name);
                        pNewNodes->name = pCurrent->name;
                        pCurrent->name = NULL;
                    }
                    /* All info and links have been transferred.  Free up
                     * remnants and leftover things. */
                    LdrfFreeTypeNodes(pCurrent);
                    pTypeTree->pCurrent = pNewNodes;
                    pReferentTree->pNodes = NULL;
                    errCode = LdrfFreeTypeTree(pReferentTree);
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfGraftReference */

LDRF_FN LdrfApplyValOverride(PLdrfTypeTree pTypeTree, PUByteArray pValMin, PUByteArray pValMax) {
    return LdrfApplyValOverrideEx(pTypeTree, pValMin, pValMax, NULL);
} /* end of: LdrfApplyValOverride */

LDRF_FN LdrfApplyValOverrideEx(PLdrfTypeTree pTypeTree, PUByteArray pValMin, PUByteArray pValMax, PUByteArray pValInvalid) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateTypeTreePointer(pTypeTree);
    TUShort typeSize, typeOffset;
    TNVTType nodeType;
    TBool hasRefs, bitfSigned;
    TUByte bitfOffset = 0, bitfSize = 0, bitfMask = 0, bitfShift = 0;

    if (!errCode) {
        if ((pValMin) || (pValMax) || (pValInvalid)) {
            errCode = LdrfScanTypeTree(pTypeTree, &typeSize, &hasRefs);

            if ((!errCode) && (hasRefs)) {
                errCode = LDRF_ERR_SEQUENCE;
            }

            while (!errCode) {
                errCode = LdrfReadTypeTreeNode(pTypeTree, &nodeType, &typeOffset, &typeSize, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

                if (errCode) {
                    if (errCode == LDRF_ERR_INCOMPLETE) {
                        errCode = LDRF_ERR_NONE;
                    } else {
                        if (errCode == LDRF_ERR_NO_MORE_NODES) {
                            errCode = LDRF_ERR_NONE;
                        }
                        break;
                    }
                }

                switch (nodeType) {
                    case NVT_TYPE_BITF:
                        errCode = LdrfGetBitfieldDetails(pTypeTree, &bitfSize, &bitfOffset, &bitfSigned, NULL, NULL, NULL, NULL, NULL);

                        if (errCode == LDRF_ERR_INCOMPLETE) {
                            errCode = LDRF_ERR_NONE;
                            bitfMask = (TUByte)((((TUShort)1) << bitfSize) - 1);
                            bitfShift = 8 - bitfOffset - bitfSize;

                            if (bitfSigned) {
                                goto SIGNED_BYTE_VALIDATION;
                            } else {
                                goto UNSIGNED_BYTE_VALIDATION;
                            }
                        }
                        break;

                    case NVT_TYPE_SIGNED_CHAR:
                    case NVT_TYPE_SIGNED_SHORT:
                    case NVT_TYPE_ENUM:
                        SIGNED_BYTE_VALIDATION:
                        {
                            TByte b;

                            if (pValMin) {
                                b = (TByte)(pValMin[typeOffset]);

                                if (nodeType == NVT_TYPE_ENUM) {
                                    pTypeTree->pCurrent->u.NVTEnum.minValid = b;
                                } else if (nodeType == NVT_TYPE_BITF) {
                                    pTypeTree->pCurrent->u.NVTBitf.minValid = (TUByte)((b >> bitfShift) & bitfMask);
                                } else {
                                    pTypeTree->pCurrent->u.NVTScalar.minValid = b;
                                }
                            }

                            if (pValMax) {
                                b = (TByte)(pValMax[typeOffset]);

                                if (nodeType == NVT_TYPE_ENUM) {
                                    pTypeTree->pCurrent->u.NVTEnum.maxValid = b;
                                } else if (nodeType == NVT_TYPE_BITF) {
                                    pTypeTree->pCurrent->u.NVTBitf.maxValid = (TUByte)((b >> bitfShift) & bitfMask);
                                } else {
                                    pTypeTree->pCurrent->u.NVTScalar.maxValid = b;
                                }
                            }

                            if (pValInvalid) {
                                b = (TByte)(pValInvalid[typeOffset]);

                                if (nodeType == NVT_TYPE_SIGNED_CHAR || nodeType == NVT_TYPE_SIGNED_SHORT) {
                                    pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                                    pTypeTree->pCurrent->u.NVTScalar.invalidVal = b;
                                } 
                            }
                        }
                        break;

                    case NVT_TYPE_UNSIGNED_CHAR:
                    case NVT_TYPE_UNSIGNED_SHORT:
                        UNSIGNED_BYTE_VALIDATION:
                        if (pValMin) {
                            if (nodeType == NVT_TYPE_BITF) {
                                TUByte b = (TUByte)((pValMin[typeOffset] >> bitfShift) & bitfMask);
                                pTypeTree->pCurrent->u.NVTBitf.minValid = b;
                            } else {
                                pTypeTree->pCurrent->u.NVTScalar.minValid = pValMin[typeOffset];
                            }
                        }

                        if (pValMax) {
                            if (nodeType == NVT_TYPE_BITF) {
                                TUByte b = (TUByte)((pValMax[typeOffset] >> bitfShift) & bitfMask);
                                pTypeTree->pCurrent->u.NVTBitf.maxValid = b;
                            } else {
                                pTypeTree->pCurrent->u.NVTScalar.maxValid = pValMax[typeOffset];
                            }
                        }

                        if (pValInvalid) {
                            if (nodeType == NVT_TYPE_UNSIGNED_CHAR || nodeType == NVT_TYPE_UNSIGNED_SHORT) {
                                pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                                pTypeTree->pCurrent->u.NVTScalar.invalidVal = pValInvalid[typeOffset];
                            } 
                        }
                        break;

                    case NVT_TYPE_SIGNED_LONG:
                    case NVT_TYPE_UNSIGNED_LONG: {
                        TUByte bufMin[2], bufMax[2], bufInvalid[2];

                        if (pValMin) {
#ifdef LCADRF_BYTE_SWAP

                            bufMin[1] = pValMin[typeOffset];
                            bufMin[0] = pValMin[typeOffset + 1];
#else

                            bufMin[0] = pValMin[typeOffset];
                            bufMin[1] = pValMin[typeOffset + 1];

#endif

                        }

                        if (pValMax) {
#ifdef LCADRF_BYTE_SWAP

                            bufMax[1] = pValMax[typeOffset];
                            bufMax[0] = pValMax[typeOffset + 1];

#else

                            bufMax[0] = pValMax[typeOffset];
                            bufMax[1] = pValMax[typeOffset + 1];

#endif

                        }

                        if (pValInvalid) {
#ifdef LCADRF_BYTE_SWAP

                            bufInvalid[1] = pValInvalid[typeOffset];
                            bufInvalid[0] = pValInvalid[typeOffset + 1];

#else

                            bufInvalid[0] = pValInvalid[typeOffset];
                            bufInvalid[1] = pValInvalid[typeOffset + 1];

#endif

                        }

                        if (nodeType == NVT_TYPE_SIGNED_LONG) {
                            TShort val;

                            if (pValMin) {
                                memcpy(&val, bufMin, sizeof(TShort));
                                pTypeTree->pCurrent->u.NVTScalar.minValid = val;
                            }

                            if (pValMax) {
                                memcpy(&val, bufMax, sizeof(TShort));
                                pTypeTree->pCurrent->u.NVTScalar.maxValid = val;
                            }

                            if (pValInvalid) {
                                memcpy(&val, bufInvalid, sizeof(TShort));
                                pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                                pTypeTree->pCurrent->u.NVTScalar.invalidVal = val;
                            }
                        } else {
                            TUShort val;

                            if (pValMin) {
                                memcpy(&val, bufMin, sizeof(TUShort));
                                pTypeTree->pCurrent->u.NVTScalar.minValid = val;
                            }

                            if (pValMax) {
                                memcpy(&val, bufMax, sizeof(TUShort));
                                pTypeTree->pCurrent->u.NVTScalar.maxValid = val;
                            }

                            if (pValInvalid) {
                                memcpy(&val, bufInvalid, sizeof(TUShort));
                                pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                                pTypeTree->pCurrent->u.NVTScalar.invalidVal = val;
                            }
                        }
                    }
                        break;

                    case NVT_TYPE_FLOAT:
                    case NVT_TYPE_SIGNED_QUAD:
                    case NVT_TYPE_UNSIGNED_QUAD: {
                        TUByte bufMin[4], bufMax[4], bufInvalid[4];

                        if (pValMin) {
#ifdef LCADRF_BYTE_SWAP

                            bufMin[3] = pValMin[typeOffset];
                            bufMin[2] = pValMin[typeOffset + 1];
                            bufMin[1] = pValMin[typeOffset + 2];
                            bufMin[0] = pValMin[typeOffset + 3];

#else

                            bufMin[0] = pValMin[typeOffset];
                            bufMin[1] = pValMin[typeOffset + 1];
                            bufMin[2] = pValMin[typeOffset + 2];
                            bufMin[3] = pValMin[typeOffset + 3];

#endif

                        }

                        if (pValMax) {
#ifdef LCADRF_BYTE_SWAP

                            bufMax[3] = pValMax[typeOffset];
                            bufMax[2] = pValMax[typeOffset + 1];
                            bufMax[1] = pValMax[typeOffset + 2];
                            bufMax[0] = pValMax[typeOffset + 3];

#else

                            bufMax[0] = pValMax[typeOffset];
                            bufMax[1] = pValMax[typeOffset + 1];
                            bufMax[2] = pValMax[typeOffset + 2];
                            bufMax[3] = pValMax[typeOffset + 3];

#endif

                        }

                        if (pValInvalid) {
#ifdef LCADRF_BYTE_SWAP

                            bufInvalid[3] = pValInvalid[typeOffset];
                            bufInvalid[2] = pValInvalid[typeOffset + 1];
                            bufInvalid[1] = pValInvalid[typeOffset + 2];
                            bufInvalid[0] = pValInvalid[typeOffset + 3];

#else

                            bufInvalid[0] = pValInvalid[typeOffset];
                            bufInvalid[1] = pValInvalid[typeOffset + 1];
                            bufInvalid[2] = pValInvalid[typeOffset + 2];
                            bufInvalid[3] = pValInvalid[typeOffset + 3];

#endif

                        }

                        if (nodeType == NVT_TYPE_FLOAT) {
                            float val;

                            if (pValMin) {
                                memcpy(&val, bufMin, sizeof(float));
                                pTypeTree->pCurrent->u.NVTFloat.minValid = val;
                            }

                            if (pValMax) {
                                memcpy(&val, bufMax, sizeof(float));
                                pTypeTree->pCurrent->u.NVTFloat.maxValid = val;
                            }
                        } else {
                            TLong val;

                            if (pValMin) {
                                memcpy(&val, bufMin, sizeof(TLong));
                                pTypeTree->pCurrent->u.NVTScalar.minValid = val;
                            }

                            if (pValMax) {
                                memcpy(&val, bufMax, sizeof(TLong));
                                pTypeTree->pCurrent->u.NVTScalar.maxValid = val;
                            }

                            if (pValInvalid) {
                                memcpy(&val, bufInvalid, sizeof(TLong));
                                pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                                pTypeTree->pCurrent->u.NVTScalar.invalidVal = val;
                            }
                        }
                    }
                        break;

                    case NVT_TYPE_DOUBLE_FLOAT: {
                        TUByte bufMin[8], bufMax[8];
                        double val;

                        if (pValMin) {
#ifdef LCADRF_BYTE_SWAP

                            bufMin[7] = pValMin[typeOffset];
                            bufMin[6] = pValMin[typeOffset + 1];
                            bufMin[5] = pValMin[typeOffset + 2];
                            bufMin[4] = pValMin[typeOffset + 3];
                            bufMin[3] = pValMin[typeOffset + 4];
                            bufMin[2] = pValMin[typeOffset + 5];
                            bufMin[1] = pValMin[typeOffset + 6];
                            bufMin[0] = pValMin[typeOffset + 7];

#else

                            bufMin[0] = pValMin[typeOffset];
                            bufMin[1] = pValMin[typeOffset + 1];
                            bufMin[2] = pValMin[typeOffset + 2];
                            bufMin[3] = pValMin[typeOffset + 3];
                            bufMin[4] = pValMin[typeOffset + 4];
                            bufMin[5] = pValMin[typeOffset + 5];
                            bufMin[6] = pValMin[typeOffset + 6];
                            bufMin[7] = pValMin[typeOffset + 7];

#endif

                        }

                        if (pValMax) {
#ifdef LCADRF_BYTE_SWAP

                            bufMax[7] = pValMax[typeOffset];
                            bufMax[6] = pValMax[typeOffset + 1];
                            bufMax[5] = pValMax[typeOffset + 2];
                            bufMax[4] = pValMax[typeOffset + 3];
                            bufMax[3] = pValMax[typeOffset + 4];
                            bufMax[2] = pValMax[typeOffset + 5];
                            bufMax[1] = pValMax[typeOffset + 6];
                            bufMax[0] = pValMax[typeOffset + 7];

#else

                            bufMax[0] = pValMax[typeOffset];
                            bufMax[1] = pValMax[typeOffset + 1];
                            bufMax[2] = pValMax[typeOffset + 2];
                            bufMax[3] = pValMax[typeOffset + 3];
                            bufMax[4] = pValMax[typeOffset + 4];
                            bufMax[5] = pValMax[typeOffset + 5];
                            bufMax[6] = pValMax[typeOffset + 6];
                            bufMax[7] = pValMax[typeOffset + 7];

#endif

                        }

                        if (pValMin) {
                            memcpy(&val, bufMin, sizeof(double));
                            pTypeTree->pCurrent->u.NVTDouble.minValid = val;
                        }

                        if (pValMax) {
                            memcpy(&val, bufMax, sizeof(double));
                            pTypeTree->pCurrent->u.NVTDouble.maxValid = val;
                        }
                    }
                        break;

                    case NVT_TYPE_SIGNED_INT64:
                    case NVT_TYPE_UNSIGNED_INT64: {
                        
                        TUByte bufMin[sizeof(__drf_int64)];
                        TUByte bufMax[sizeof(__drf_int64)];
                        TUByte bufInvalid[sizeof(__drf_int64)];
                        if (pValMin) {
                            LdrfCopyAndSwapScalar(bufMin, &pValMin[typeOffset], sizeof(bufMin));
                        }

                        if (pValMax) {
                            LdrfCopyAndSwapScalar(bufMax, &pValMax[typeOffset], sizeof(bufMax));
                        }

                        if (pValInvalid) {
                            LdrfCopyAndSwapScalar(bufInvalid, &pValInvalid[typeOffset], sizeof(bufInvalid));
                        }

                        if (pValMin) {
                            memcpy(&pTypeTree->pCurrent->u.NVTScalar64.minValid, bufMin, 
                                    sizeof(pTypeTree->pCurrent->u.NVTScalar64.minValid));
                        }

                        if (pValMax) {
                            memcpy(&pTypeTree->pCurrent->u.NVTScalar64.maxValid, bufMax, 
                                    sizeof(pTypeTree->pCurrent->u.NVTScalar64.maxValid));
                        }

                        if (pValInvalid) {
                            memcpy(&pTypeTree->pCurrent->u.NVTScalar64.invalidVal, bufInvalid, 
                                    sizeof(pTypeTree->pCurrent->u.NVTScalar64.invalidVal));
                            pTypeTree->pCurrent->u.NVTScalar.invalidValuePresent = T;
                        }
                    }
                        break;

                    case NVT_TYPE_ARRAY:
                        errCode = LdrfGetArrayDetails(pTypeTree, F, NULL);

                        if (errCode == LDRF_ERR_INCOMPLETE) {
                            errCode = LDRF_ERR_NONE;
                        }
                        break;

                    default:
                        /* Adding this clause here prevents some compilers
                         * from creating warning messages */
                        ;
                } /* end of: switch (nodeType) */
            }     /* end of: while (!errCode) */
        }
    }

    return errCode;
} /* end of: LdrfApplyValOverrideEx */
